from pandas import DataFrame, json_normalize, set_option, Series

set_option('display.max_columns', None)
set_option('display.width', 1000)
set_option('display.colheader_justify', 'center')
set_option('display.precision', 3)
import numpy


class SampleSet:
    """
       An object that stores data of type SampleSet. To understand the significance of each column of the DataFrame. Please visit "SampleSet" in [PGS Catalog Documentation](https://www.pgscatalog.org/rest/) for details.

       Attributes:
            raw_data: list. Convert from obtained JSON data
            sample_sets: DataFrame. It only exists if the parameter mode of constructor is Fat.
            samples: DataFrame. It only exists if the parameter mode of constructor is Fat.
            cohorts: DataFrame. It only exists if the parameter mode of constructor is Fat.
            mode: Fat or Thin. Specifies the mode of the returned object.

       ```python
       from pandaspgs.get_sample_set import get_sample_sets

       ch = get_sample_sets(pss_id='PSS000001')
       ch
       ch.raw_data
       ch.mode
       ch.sample_sets
       ch.samples
       ch.cohorts
       ```
       Subset object s by either identifier or position
       ```python
       all_df = get_sample_sets()
       all_df[0].sample_sets
       all_df[0:3].sample_sets
       all_df['PSS000001'].sample_sets
       all_df[('PSS000001','PSS000002','PSS000003')].sample_sets
       ```
       Objects can be manipulated like sets in the mathematical sense.
       ```python
       one = get_sample_sets(pss_id='PSS000001')
       two = get_sample_sets(pss_id='PSS000002')
       three = get_sample_sets(pss_id='PSS000003')
       one_and_two = one+two
       two_and_three = two+three
       only_one = one_and_two - two_and_three
       only_two = one_and_two & two_and_three
       one_and_two_and_three = one_and_two | two_and_three
       one_and_three = one_and_two ^ two_and_three
       ```
       """
    def __init__(self, data: list = [], mode: str = "Fat"):
        """
        Args:
            data: Raw JSON data.
            mode: Fat or Thin. Specifies the mode of the object.
        """
        if data is None:
            data = []
        if mode not in ['Thin', "Fat"]:
            raise Exception("Mode must be Fat or Thin")
        self.raw_data = data
        self.mode = mode
        if mode == "Thin":
            return
        if data is None or len(data) == 0:
            self.sample_sets = DataFrame(
                columns=['id'])
            self.samples = DataFrame(
                columns=['id', 'sample_set_id', 'sample_number', 'sample_cases', 'sample_controls',
                         'sample_percent_male', 'sample_age.estimate_type', 'sample_age.estimate',
                         'sample_age.interval.type', 'sample_age.interval.lower', 'sample_age.interval.upper',
                         'sample_age.variability_type', 'sample_age.variability', 'sample_age.unit',
                         'phenotyping_free', 'followup_time.estimate_type', 'followup_time.estimate',
                         'followup_time.interval.type', 'followup_time.interval.lower', 'followup_time.interval.upper',
                         'followup_time.variability_type', 'followup_time.variability', 'followup_time.unit',
                         'ancestry_broad', 'ancestry_free', 'ancestry_country', 'ancestry_additional',
                         'source_GWAS_catalog', 'source_PMID', 'source_DOI', 'cohorts_additional'])
            self.cohorts = DataFrame(columns=['sample_set_id', 'sample_id', 'name_short', 'name_full', 'name_others'])
            return
        datas = json_normalize(data=data, max_level=1)
        datas['samples'] = datas['samples'].map(lambda x: x == [])
        self.sample_sets = json_normalize(data=data, max_level=1).drop(
            columns=['samples'])
        if not datas['samples'].all():
            self.samples = json_normalize(data=data, record_path=['samples'], meta=['id'])
            self.samples['sample_set_id'] = self.samples['id']
            self.samples['id'] = Series(data=range(0, len(self.samples)))
            cohort = self.samples[['id', 'sample_set_id', 'cohorts']].copy()
            self.samples = self.samples.drop(columns=['cohorts'])
            cohort['sample_id'] = cohort['id']
            cohort['cohorts'] = cohort['cohorts'].apply(lambda x: x if len(x) > 0 else numpy.nan)
            cohort = cohort.dropna()
            cohort = cohort.explode('cohorts')
            cohort[['name_short', 'name_full', 'name_others']] = cohort['cohorts'].apply(
                lambda x: Series(data=[x['name_short'], x['name_full'], x['name_others']]))
            cohort = cohort.drop(columns=['id', 'cohorts'])
            self.cohorts = cohort

        else:
            self.samples = DataFrame(
                columns=['id', 'sample_set_id', 'sample_number', 'sample_cases', 'sample_controls',
                         'sample_percent_male', 'sample_age.estimate_type', 'sample_age.estimate',
                         'sample_age.interval.type', 'sample_age.interval.lower', 'sample_age.interval.upper',
                         'sample_age.variability_type', 'sample_age.variability', 'sample_age.unit',
                         'phenotyping_free', 'followup_time.estimate_type', 'followup_time.estimate',
                         'followup_time.interval.type', 'followup_time.interval.lower', 'followup_time.interval.upper',
                         'followup_time.variability_type', 'followup_time.variability', 'followup_time.unit',
                         'ancestry_broad', 'ancestry_free', 'ancestry_country', 'ancestry_additional',
                         'source_GWAS_catalog', 'source_PMID', 'source_DOI', 'cohorts_additional'])
            self.cohorts = DataFrame(columns=['sample_set_id', 'sample_id', 'name_short', 'name_full', 'name_others'])
        if 'publication' in self.sample_sets.columns:
            self.sample_sets = self.sample_sets.drop(columns=['pubication'])
            self.sample_sets = self.sample_sets.reindex(
                columns=self.sample_sets.columns.tolist() + ['publication.title',
                                                             'publication.doi', 'publication.PMID',
                                                             'publication.journal',
                                                             'publication.firstauthor',
                                                             'publication.date_publication'])
        if 'sample_age' in self.samples.columns:
            self.samples = self.samples.drop(columns=['sample_age'])
            if 'sample_age.estimate_type' not in self.samples.columns:
                self.samples = self.samples.reindex(
                    columns=self.samples.columns.tolist() + ['sample_age.estimate_type', 'sample_age.estimate',
                                                             'sample_age.interval.type', 'sample_age.interval.lower',
                                                             'sample_age.interval.upper',
                                                             'sample_age.variability_type', 'sample_age.variability',
                                                             'sample_age.unit']
            )
        if 'followup_time' in self.samples.columns:
            self.samples = self.samples.drop(columns=['followup_time'])
            if 'followup_time.estimate_type' not in self.samples.columns:
                self.samples = self.samples.reindex(
                    columns=self.samples.columns.tolist() + ['followup_time.estimate_type', 'followup_time.estimate',
                                                             'followup_time.interval.type',
                                                             'followup_time.interval.lower',
                                                             'followup_time.interval.upper',
                                                             'followup_time.variability_type',
                                                             'followup_time.variability',
                                                             'followup_time.unit'])
        return

    def __str__(self):
        if self.mode == 'Fat':
            return ("SampleSet is running in fat mode. It has 3 DataFrames with hierarchical dependencies.\n"
                    "sample_sets:"
                    "%d rows\n|\n -samples: %d rows\n  |\n   -cohorts: %d rows" % (
                        len(self.sample_sets), len(self.samples), len(self.cohorts)))
        if self.mode == 'Thin':
            return ('SampleSet is running in thin mode. It has 1 list that contains the raw data.\nraw_data: a list of '
                    'size %d.' % len(self.raw_data))

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        if isinstance(item, str) or isinstance(item, int):
            arr = [item]
        elif isinstance(item, list) or isinstance(item, tuple) or isinstance(item, range):
            arr = item
        elif isinstance(item, slice):
            start, stop, step = item.indices(len(self))
            arr = list(range(start, stop, step))
        else:
            raise TypeError('Invalid argument type：{}'.format(type(item)))
        raw_data = self.raw_data
        raw_data_dict = {}
        for j in raw_data:
            raw_data_dict[j['id']] = j
        sub_set = []
        for i in arr:
            if isinstance(i, str):
                sub_set.append(raw_data_dict[i])
            elif isinstance(i, int):
                sub_set.append(raw_data[i])
            else:
                raise TypeError('Invalid item type: {}'.format(type(i)))
        return SampleSet(sub_set, self.mode)

    def __add__(self, other):
        if self.mode == other.mode:
            return SampleSet(self.raw_data + other.raw_data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __sub__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set - other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return SampleSet(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __and__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set & other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return SampleSet(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __or__(self, other):
        if self.mode == other.mode:
            and_dict = {}
            for i in self.raw_data:
                and_dict[i['id']] = i
            for j in other.raw_data:
                and_dict[j['id']] = j
            data = list(and_dict.values())
            return SampleSet(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __xor__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            and_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                and_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
                and_dict[j['id']] = j
            sub_key = self_key_set ^ other_key_set
            data = []
            for k in sub_key:
                data.append(and_dict[k])
            return SampleSet(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __eq__(self, other):
        if self is None or other is None:
            return self is None and other is None
        return self.raw_data == other.raw_data and self.mode == other.mode

    def __len__(self):
        return len(self.raw_data)



from pandaspgs.get_publication import get_publications
from pandaspgs.client import clear_cache

def test_get_publications():
    filter_by_id = get_publications(pgp_id='PGP000001')
    assert len(filter_by_id) == 1
    filter_by_pgs_id = get_publications(pgs_id='PGS000001')
    assert len(filter_by_pgs_id) == 1
    filter_by_pmid = get_publications(pmid=25855707)
    assert len(filter_by_pmid) == 1
    filter_by_author = get_publications(author='Mavaddat')
    assert len(filter_by_author) == 6
    filter_by_all = get_publications(pgs_id='PGS000001', pgp_id='PGP000001', pmid=25855707, author='Mavaddat')
    assert len(filter_by_all) == 1
    filter_by_none = get_publications()
    clear_cache('Publication')
    clear_cache('All')
    assert len(filter_by_none) == 537
    assert len(filter_by_none ^ filter_by_id) == 536
    assert len(filter_by_none[range(2)]) == 2
    assert len(filter_by_none[1:3]) == 2
    assert filter_by_none['PGP000001'] == filter_by_id
    assert len(filter_by_id + filter_by_pgs_id) == 2
    assert len(filter_by_none - filter_by_id) == 536
    assert len(filter_by_none & filter_by_id) == 1
    assert len(filter_by_id | filter_by_pgs_id) == 1
    assert filter_by_id == filter_by_pgs_id
    assert len(filter_by_none[0:506] | filter_by_none[506]) == 507




from pandaspgs.client import get_sample_set
from pandaspgs.sample_set import SampleSet


def get_sample_sets(pss_id: str = None, pgs_id: str = None, pgp_id: str = None, pmid: int = None, cached: bool = True,
                    mode: str = 'Fat') \
        -> SampleSet:
    """
    Get SampleSet data from the server.

    Args:
        pss_id: PGS Sample Set ID (PSS)
        pgs_id: Polygenic Score ID (PGS)
        pgp_id: PGS Catalog Publication ID (PGP)
        pmid:PubMed ID (without the prefix "PMID:")
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A SampleSet object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_sample_set import get_sample_sets

    ch = get_sample_sets(pss_id='PSS000001')
    ```
    """
    if pss_id is None and pgs_id is None and pgp_id is None and pmid is None:
        return SampleSet(get_sample_set('https://www.pgscatalog.org/rest/sample_set/all', cached=cached), mode)
    by_id = None
    by_other = None
    if pss_id is not None:
        by_id = get_sample_set('https://www.pgscatalog.org/rest/sample_set/%s' % pss_id, cached=cached)
    if pgs_id is not None or pmid is not None or pgp_id is not None:
        query_str = []
        if pgs_id is not None:
            query_str.append('pgs_id=%s' % pgs_id)
        if pmid is not None:
            query_str.append('pmid=%d' % pmid)
        if pgp_id is not None:
            query_str.append('pgp_id=%s' % pgp_id)
        by_other = get_sample_set('https://www.pgscatalog.org/rest/sample_set/search?%s' % '&'.join(query_str), cached=cached)
    if pss_id is None:
        return SampleSet(by_other, mode)
    if pgs_id is None and pmid is None and pgp_id is None:
        return SampleSet(by_id, mode)
    other_set = set()
    id_dict = {}
    for single in by_id:
        id_dict[single['id']] = single
    ppm_id_set = id_dict.keys()
    for single in by_other:
        other_set.add(single['id'])
    intersection = ppm_id_set & other_set
    result = []
    for id in intersection:
        result.append(id_dict[id])
    return SampleSet(result, mode)





from pandaspgs.client import get_ancestry_category
from pandaspgs.ancestry_category import AncestryCategory


def get_ancestry_categories(cached: bool = True, mode: str = 'Fat') -> AncestryCategory:
    """
    Get AncestryCategory data from the server.

    Args:
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.

    Returns:
        A AncestryCategory object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_ancestry_category import get_ancestry_categories

    ch = get_ancestry_categories()
    ```
    """
    return AncestryCategory(get_ancestry_category('https://www.pgscatalog.org/rest/ancestry_categories/',
                                                  cached=cached), mode)





from pandaspgs.cohort import Cohort
from pandaspgs.score import Score
from pandaspgs.sample_set import SampleSet
from pandaspgs.release import Release
from pandaspgs.trait import Trait
from pandaspgs.publication import Publication
from pandaspgs.ancestry_category import AncestryCategory
from pandaspgs.browser import open_sample_set_in_pgs_catalog, open_publication_in_pgs_catalog, \
    open_score_in_pgs_catalog, open_trait_in_pgs_catalog, open_in_dbsnp, open_in_pubmed
from pandaspgs.performance import PerformanceMetric
from pandaspgs.trait_category import TraitCategory
from pandaspgs.get_performance import get_performances
from pandaspgs.get_ancestry_category import get_ancestry_categories
from pandaspgs.get_trait import get_traits, get_trait_categories, get_child_traits
from pandaspgs.get_score import get_scores
from pandaspgs.get_cohort import get_cohorts
from pandaspgs.get_publication import get_publications
from pandaspgs.get_release import get_releases
from pandaspgs.get_sample_set import get_sample_sets
from pandaspgs.set_operation import set_equal, set_xor, set_diff, intersect, union, bind
from pandaspgs.file_operation import read_scoring_file, write_csv, write_xlsx
from pandaspgs.client import clear_cache

__all__ = ['Cohort', 'Score', 'SampleSet', 'Release', 'Trait', 'Publication', 'AncestryCategory',
           'open_sample_set_in_pgs_catalog', 'open_publication_in_pgs_catalog', 'open_score_in_pgs_catalog',
           'open_trait_in_pgs_catalog', 'open_in_dbsnp', 'open_in_pubmed', 'PerformanceMetric',
           'TraitCategory', 'get_performances', 'get_ancestry_categories', 'get_traits', 'get_trait_categories',
           'get_child_traits', 'get_scores', 'get_cohorts', 'get_publications', 'get_releases', 'get_sample_sets',
           'set_equal', 'set_xor', 'set_diff', 'intersect', 'union', 'bind', 'read_scoring_file', 'write_csv',
           'write_xlsx']



from pandas import DataFrame, json_normalize, set_option
from pandas import concat

set_option('display.max_columns', None)
set_option('display.width', 1000)
set_option('display.colheader_justify', 'center')
set_option('display.precision', 3)


class Publication:
    """
       An object that stores data of type Publication. To understand the significance of each column of the DataFrame. Please visit "Publication" in [PGS Catalog Documentation](https://www.pgscatalog.org/rest/) for details.

       Attributes:
            raw_data: list. Convert from obtained JSON data
            publications: DataFrame. It only exists if the parameter mode of constructor is Fat.
            associated_pgs_ids: DataFrame. It only exists if the parameter mode of constructor is Fat.
            mode: Fat or Thin. Specifies the mode of the returned object.

       ```python
       from pandaspgs.get_publication import get_publications

       ch = get_publications(pgs_id='PGS000001')
       ch
       ch.raw_data
       ch.mode
       ch.publications
       ch.associated_pgs_ids
       ```
       Subset object s by either identifier or position
       ```python
       all_df = get_publications()
       all_df[0].publications
       all_df[0:3].publications
       all_df['PGP000001'].publications
       all_df[('PGP000001','PGP000002','PGP000003')].publications
       ```
       Objects can be manipulated like sets in the mathematical sense.
       ```python
       one = get_cohorts(ppm_id='PPM000001')
       two = get_cohorts(ppm_id='PPM000002')
       three = get_cohorts(ppm_id='PPM000003')
       one_and_two = one+two
       two_and_three = two+three
       only_one = one_and_two - two_and_three
       only_two = one_and_two & two_and_three
       one_and_two_and_three = one_and_two | two_and_three
       one_and_three = one_and_two ^ two_and_three
       ```
       """
    def __init__(self, data: list = [], mode: str = "Fat"):
        """
        Args:
            data: Raw JSON data.
            mode: Fat or Thin. Specifies the mode of the object.
        """
        if data is None:
            data = []
        if mode not in ['Thin', "Fat"]:
            raise Exception("Mode must be Fat or Thin")
        self.raw_data = data
        self.mode = mode
        if mode == "Thin":
            return
        if data is None or len(data) == 0:
            self.publications = DataFrame(
                columns=['id', 'title', 'doi', 'PMID', 'journal', 'firstauthor', 'date_publication', 'date_release'
                    , 'authors'])
            self.associated_pgs_ids = DataFrame(
                columns=['publication_id', 'associated_pgs_id', 'stage'])
            return
        datas = json_normalize(data=data, max_level=1).drop(columns=['id', 'title', 'doi', 'PMID', 'journal'
            , 'firstauthor', 'date_publication', 'date_release', 'authors'])
        datas['associated_pgs_ids.development'] = datas['associated_pgs_ids.development'].map(lambda x: x == [])
        datas['associated_pgs_ids.evaluation'] = datas['associated_pgs_ids.evaluation'].map(lambda x: x == [])
        self.publications = json_normalize(data=data, max_level=1).drop(
            columns=['associated_pgs_ids.development', 'associated_pgs_ids.evaluation'])
        if not datas['associated_pgs_ids.development'].all() or not datas['associated_pgs_ids.evaluation'].all():
            dva = json_normalize(data=data, record_path=['associated_pgs_ids', 'development'], meta=['id'])
            eva = json_normalize(data=data, record_path=['associated_pgs_ids', 'evaluation'], meta=['id'])
            dva['stage'] = 'development'
            eva['stage'] = 'evaluation'
            self.associated_pgs_ids = concat([dva, eva])
            self.associated_pgs_ids.columns = ['associated_pgs_id', 'publication_id', 'stage']
        else:
            self.associated_pgs_ids = DataFrame(
                columns=['associated_pgs_id', 'publication_id', 'stage'])

        return

    def __str__(self):
        if self.mode == 'Fat':
            return ("Publication is running in fat mode. It has 2 DataFrames with hierarchical dependencies.\n"
                    "publications: %d rows\n|\n -associated_pgs_ids: %d rows" % (len(self.publications),
                                                                                 len(self.associated_pgs_ids)))
        if self.mode == 'Thin':
            return ('Publication is running in thin mode. It has 1 list that contains the raw data.\nraw_data: a list '
                    'of size %d.') % len(self.raw_data)

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        if isinstance(item, str) or isinstance(item, int):
            arr = [item]
        elif isinstance(item, list) or isinstance(item, tuple) or isinstance(item, range):
            arr = item
        elif isinstance(item, slice):
            start, stop, step = item.indices(len(self))
            arr = list(range(start, stop, step))
        else:
            raise TypeError('Invalid argument type：{}'.format(type(item)))
        raw_data = self.raw_data
        raw_data_dict = {}
        for j in raw_data:
            raw_data_dict[j['id']] = j
        sub_set = []
        for i in arr:
            if isinstance(i, str):
                sub_set.append(raw_data_dict[i])
            elif isinstance(i, int):
                sub_set.append(raw_data[i])
            else:
                raise TypeError('Invalid item type: {}'.format(type(i)))
        return Publication(sub_set, self.mode)

    def __add__(self, other):
        if self.mode == other.mode:
            return Publication(self.raw_data + other.raw_data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __sub__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set - other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Publication(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __and__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set & other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Publication(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __or__(self, other):
        if self.mode == other.mode:
            and_dict = {}
            for i in self.raw_data:
                and_dict[i['id']] = i
            for j in other.raw_data:
                and_dict[j['id']] = j
            data = list(and_dict.values())
            return Publication(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __xor__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            and_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                and_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
                and_dict[j['id']] = j
            sub_key = self_key_set ^ other_key_set
            data = []
            for k in sub_key:
                data.append(and_dict[k])
            return Publication(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __eq__(self, other):
        if self is None or other is None:
            return self is None and other is None
        return self.raw_data == other.raw_data and self.mode == other.mode

    def __len__(self):
        return len(self.raw_data)



from pandaspgs.get_sample_set import get_sample_sets


def test_get_sample_sets():
    filter_by_id = get_sample_sets()
    assert len(filter_by_id) == 9112
    assert len(filter_by_id ^ filter_by_id[0]) == 9111
    assert len(filter_by_id[range(2)]) == 2
    assert len(filter_by_id[1:3]) == 2
    assert len(filter_by_id['PSS011331']) == 1
    assert len(filter_by_id[0] + filter_by_id[1]) == 2
    assert len(filter_by_id - filter_by_id[1]) == 9111
    assert len(filter_by_id[0] & filter_by_id) == 1
    assert len(filter_by_id | filter_by_id[0]) == 9112
    assert len(filter_by_id[0:506] | filter_by_id[506]) == 507





import webbrowser


def open_score_in_pgs_catalog(pgs_id: str) -> bool:
    """
    This function launches the web browser and opens a tab for each identifier on the PGS Catalog web graphical user interface: https://www.pgscatalog.org/.

    Args:
        pgs_id: This argument indicates the type of the identifiers passed in identifier

    Returns:
        Returns TRUE if successful, or FALSE otherwise. But note that this function is run for its side effect.
    ```python
    from pandaspgs.browser import open_score_in_pgs_catalog
    open_score_in_pgs_catalog('PGS000001')
    ```

    """
    return webbrowser.open_new_tab('https://www.pgscatalog.org/score/%s' % pgs_id)


def open_publication_in_pgs_catalog(pgp_id: str) -> bool:
    """
    This function launches the web browser and opens a tab for each identifier on the PGS Catalog web graphical user interface: https://www.pgscatalog.org/.

    Args:
        pgp_id: This argument indicates the type of the identifiers passed in identifier

    Returns:
        Returns TRUE if successful, or FALSE otherwise. But note that this function is run for its side effect.
    ```python
    from pandaspgs.browser import open_publication_in_pgs_catalog
    open_publication_in_pgs_catalog('PGP000001')
    ```

    """
    return webbrowser.open_new_tab('https://www.pgscatalog.org/publication/%s' % pgp_id)


def open_sample_set_in_pgs_catalog(pss_id: str) -> bool:
    """
    This function launches the web browser and opens a tab for each identifier on the PGS Catalog web graphical user interface: https://www.pgscatalog.org/.

    Args:
        pss_id: This argument indicates the type of the identifiers passed in identifier

    Returns:
        Returns TRUE if successful, or FALSE otherwise. But note that this function is run for its side effect.
    ```python
    from pandaspgs.browser import open_sample_set_in_pgs_catalog
    open_sample_set_in_pgs_catalog('PSS000001')
    ```

    """
    return webbrowser.open_new_tab('https://www.pgscatalog.org/sampleset/%s' % pss_id)


def open_trait_in_pgs_catalog(efo_id: str) -> bool:
    """
    This function launches the web browser and opens a tab for each identifier on the PGS Catalog web graphical user interface: https://www.pgscatalog.org/.

    Args:
        efo_id: This argument indicates the type of the identifiers passed in identifier

    Returns:
        Returns TRUE if successful, or FALSE otherwise. But note that this function is run for its side effect.
    ```python
    from pandaspgs.browser import open_trait_in_pgs_catalog
    open_trait_in_pgs_catalog('EFO_0001645')
    ```

    """
    return webbrowser.open_new_tab('https://www.pgscatalog.org/trait/%s' % efo_id)


def open_in_pubmed(pubmed_id: str) -> bool:
    """
    This function launches the web browser and opens a tab for each PubMed citation.

    Args:
        pubmed_id: A PubMed identifier, either a character or an integer vector.

    Returns:
        Returns TRUE if successful. Note however that this function is run for its side effect
    ```python
    from pandaspgs.browser import open_in_pubmed
    open_in_pubmed('26301688')
    ```

    """
    return webbrowser.open_new_tab('https://www.ncbi.nlm.nih.gov/pubmed/%s' % pubmed_id)


def open_in_dbsnp(variant_id: str) -> bool:
    """
    This function launches the web browser at dbSNP and opens a tab for each SNP identifier.

    Args:
        variant_id: A variant identifier, a character vector.

    Returns:
        Returns TRUE if successful. Note however that this function is run for its side effect.
    ```python
    from pandaspgs.browser import open_in_dbsnp
    open_in_dbsnp('rs56261590')
    ```

    """
    return webbrowser.open_new_tab('https://www.ncbi.nlm.nih.gov/snp/%s' % variant_id)




from pandaspgs import get_scores
from pandaspgs import clear_cache


def test_get_scores():
    filter_by_id = get_scores(pgs_id='PGS000001')
    assert len(filter_by_id) == 1
    filter_by_pgp_id = get_scores(pgp_id='PGP000001')
    assert len(filter_by_pgp_id) == 3
    filter_by_pmid = get_scores(pmid=25855707)
    assert len(filter_by_pmid) == 3
    filter_by_trait_id = get_scores(trait_id='EFO_1000649')
    assert len(filter_by_trait_id) == 6
    filter_by_all = get_scores(pgs_id='PGS000002', pgp_id='PGP000001', pmid=25855707, trait_id='EFO_1000649')
    assert len(filter_by_all) == 1
    filter_by_id = get_scores()
    assert len(filter_by_id) == 4650
    assert len(filter_by_id ^ filter_by_id[0]) == 4649
    assert len(filter_by_id[range(2)]) == 2
    assert len(filter_by_id[1:3]) == 2
    assert len(filter_by_id['PGS000001']) == 1
    assert len(filter_by_id[0] + filter_by_id[1]) == 2
    assert len(filter_by_id - filter_by_id[1]) == 4649
    assert len(filter_by_id[0] & filter_by_id) == 1
    assert len(filter_by_id | filter_by_id[0]) == 4650
    assert len(filter_by_id[0:506] | filter_by_id[506]) == 507




import os
import sys
import re

import requests
from pandas import DataFrame, read_table, Series
from requests.adapters import HTTPAdapter

from pandaspgs import get_scores, AncestryCategory, Cohort, PerformanceMetric, Publication, Release, SampleSet, Score, \
    Trait, TraitCategory

s = requests.Session()
s.mount('https://', HTTPAdapter(max_retries=5))
home_path = os.path.expanduser('~') + os.sep + 'pandaspgs_home'


def write_csv(path: str,
              o: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> None:
    """
    Create a directory and write the attributes of pandasPGS objects to the corresponding CSV files.

    Args:
        path: The directory that needs to be created.
        o: pandasPGS object.

    Returns:
        None

    ```Python
    from pandaspgs import *
    import os
    home_path = os.path.expanduser('~') + os.sep + 'pandaspgs_home'
    ancestry = get_ancestry_categories()
    write_csv(home_path + os.sep + 'ancestry', ancestry)
    ```

    """
    os.mkdir(path)
    if type(o) is AncestryCategory and o.mode == 'Fat':
        o.ancestry_categories.to_csv(path_or_buf=path + os.sep + 'ancestry_categories.csv', index=False, mode='x')
        o.categories.to_csv(path_or_buf=path + os.sep + 'categories.csv', index=False, mode='x')
    elif type(o) is Cohort and o.mode == 'Fat':
        o.cohorts.to_csv(path_or_buf=path + os.sep + 'cohorts.csv', index=False, mode='x')
        o.associated_pgs_ids.to_csv(path_or_buf=path + os.sep + 'associated_pgs_ids.csv', index=False, mode='x')
    elif type(o) is PerformanceMetric and o.mode == 'Fat':
        o.performance_metrics.to_csv(path_or_buf=path + os.sep + 'performance_metrics.csv', index=False, mode='x')
        o.samples.to_csv(path_or_buf=path + os.sep + 'samples.csv', index=False, mode='x')
        o.cohorts.to_csv(path_or_buf=path + os.sep + 'cohorts.csv', index=False, mode='x')
        o.effect_sizes.to_csv(path_or_buf=path + os.sep + 'effect_sizes.csv', index=False, mode='x')
        o.class_acc.to_csv(path_or_buf=path + os.sep + 'class_acc.csv', index=False, mode='x')
        o.othermetrics.to_csv(path_or_buf=path + os.sep + 'othermetrics.csv', index=False, mode='x')
    elif type(o) is Publication and o.mode == 'Fat':
        o.publications.to_csv(path_or_buf=path + os.sep + 'publications.csv', index=False, mode='x')
        o.associated_pgs_ids.to_csv(path_or_buf=path + os.sep + 'associated_pgs_ids.csv', index=False, mode='x')
    elif type(o) is Release and o.mode == 'Fat':
        o.releases.to_csv(path_or_buf=path + os.sep + 'releases.csv', index=False, mode='x')
        o.released_performance_ids.to_csv(path_or_buf=path + os.sep + 'released_performance_ids.csv', index=False,
                                          mode='x')
        o.released_publication_ids.to_csv(path_or_buf=path + os.sep + 'released_publication_ids.csv', index=False,
                                          mode='x')
        o.released_score_ids.to_csv(path_or_buf=path + os.sep + 'released_score_ids.csv', index=False, mode='x')
        o.released_new_trait_ids.to_csv(path_or_buf=path + os.sep + 'released_new_trait_ids.csv', index=False, mode='x')
    elif type(o) is SampleSet and o.mode == 'Fat':
        o.sample_sets.to_csv(path_or_buf=path + os.sep + 'sample_sets.csv', index=False, mode='x')
        o.samples.to_csv(path_or_buf=path + os.sep + 'samples.csv', index=False, mode='x')
        o.cohorts.to_csv(path_or_buf=path + os.sep + 'cohorts.csv', index=False, mode='x')
    elif type(o) is Score and o.mode == 'Fat':
        o.scores.to_csv(path_or_buf=path + os.sep + 'scores.csv', index=False, mode='x')
        o.samples_variants.to_csv(path_or_buf=path + os.sep + 'samples_variants.csv', index=False, mode='x')
        o.samples_variants_cohorts.to_csv(path_or_buf=path + os.sep + 'samples_variants_cohorts.csv', index=False,
                                          mode='x')
        o.trait_efo.to_csv(path_or_buf=path + os.sep + 'trait_efo.csv', index=False, mode='x')
        o.samples_training.to_csv(path_or_buf=path + os.sep + 'samples_training.csv', index=False, mode='x')
        o.samples_training_cohorts.to_csv(path_or_buf=path + os.sep + 'samples_training_cohorts.csv', index=False,
                                          mode='x')
        o.ancestry_distribution.to_csv(path_or_buf=path + os.sep + 'ancestry_distribution.csv', index=False, mode='x')
    elif type(o) is Trait and o.mode == 'Fat':
        o.traits.to_csv(path_or_buf=path + os.sep + 'traits.csv', index=False, mode='x')
        o.trait_categories.to_csv(path_or_buf=path + os.sep + 'trait_categories.csv', index=False, mode='x')
        o.trait_synonyms.to_csv(path_or_buf=path + os.sep + 'trait_synonyms.csv', index=False, mode='x')
        o.trait_mapped_terms.to_csv(path_or_buf=path + os.sep + 'trait_mapped_terms.csv', index=False, mode='x')
        o.associated_pgs_ids.to_csv(path_or_buf=path + os.sep + 'associated_pgs_ids.csv', index=False, mode='x')
        o.child_associated_pgs_ids.to_csv(path_or_buf=path + os.sep + 'child_associated_pgs_ids.csv', index=False,
                                          mode='x')
    elif type(o) is TraitCategory and o.mode == 'Fat':
        o.efotraits.to_csv(path_or_buf=path + os.sep + 'efotraits.csv', index=False, mode='x')
        o.trait_categories.to_csv(path_or_buf=path + os.sep + 'trait_categories.csv', index=False, mode='x')
    else:
        raise Exception(
            "Failed to convert to CSV files. "
            "Please check whether the type of the input object is supported by pandasPGS, "
            "and check whether the mode attribute of the input object is 'Fat'.")


def write_xlsx(path: str,
                o: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> None:
    """
    Create a directory and write the attributes of pandasPGS objects to the corresponding EXCEL files.

    Args:
        path: The directory that needs to be created.
        o: pandasPGS object.

    Returns:
        None

    ```Python
    from pandaspgs import *
    import os
    home_path = os.path.expanduser('~') + os.sep + 'pandaspgs_home'
    ancestry = get_ancestry_categories()
    write_xlsx(home_path + os.sep + 'ancestry', ancestry)
    ```

    """
    os.mkdir(path)
    if type(o) is AncestryCategory and o.mode == 'Fat':
        o.ancestry_categories.to_excel(path + os.sep + 'ancestry_categories.xlsx', index=False)
        o.categories.to_excel(path + os.sep + 'categories.xlsx', index=False)
    elif type(o) is Cohort and o.mode == 'Fat':
        o.cohorts.to_excel(path + os.sep + 'cohorts.xlsx', index=False)
        o.associated_pgs_ids.to_excel(path + os.sep + 'associated_pgs_ids.xlsx', index=False)
    elif type(o) is PerformanceMetric and o.mode == 'Fat':
        o.performance_metrics.to_excel(path + os.sep + 'performance_metrics.xlsx', index=False)
        o.samples.to_excel(path + os.sep + 'samples.xlsx', index=False)
        o.cohorts.to_excel(path + os.sep + 'cohorts.xlsx', index=False)
        o.effect_sizes.to_excel(path + os.sep + 'effect_sizes.xlsx', index=False)
        o.class_acc.to_excel(path + os.sep + 'class_acc.xlsx', index=False)
        o.othermetrics.to_excel(path + os.sep + 'othermetrics.xlsx', index=False)
    elif type(o) is Publication and o.mode == 'Fat':
        o.publications.to_excel(path + os.sep + 'publications.xlsx', index=False)
        o.associated_pgs_ids.to_excel(path + os.sep + 'associated_pgs_ids.xlsx', index=False)
    elif type(o) is Release and o.mode == 'Fat':
        o.releases.to_excel(path + os.sep + 'releases.xlsx', index=False)
        o.released_performance_ids.to_excel(path + os.sep + 'released_performance_ids.xlsx', index=False)
        o.released_publication_ids.to_excel(path + os.sep + 'released_publication_ids.xlsx', index=False)
        o.released_score_ids.to_excel(path + os.sep + 'released_score_ids.xlsx', index=False)
        o.released_new_trait_ids.to_excel(path + os.sep + 'released_new_trait_ids.xlsx', index=False)
    elif type(o) is SampleSet and o.mode == 'Fat':
        o.sample_sets.to_excel(path + os.sep + 'sample_sets.xlsx', index=False)
        o.samples.to_excel(path + os.sep + 'samples.xlsx', index=False)
        o.cohorts.to_excel(path + os.sep + 'cohorts.xlsx', index=False)
    elif type(o) is Score and o.mode == 'Fat':
        o.scores.to_excel(path + os.sep + 'scores.xlsx', index=False)
        o.samples_variants.to_excel(path + os.sep + 'samples_variants.xlsx', index=False)
        o.samples_variants_cohorts.to_excel(path + os.sep + 'samples_variants_cohorts.xlsx', index=False)
        o.trait_efo.to_excel(path + os.sep + 'trait_efo.xlsx', index=False)
        o.samples_training.to_excel(path + os.sep + 'samples_training.xlsx', index=False)
        o.samples_training_cohorts.to_excel(path + os.sep + 'samples_training_cohorts.xlsx', index=False)
        o.ancestry_distribution.to_excel(path + os.sep + 'ancestry_distribution.xlsx', index=False)
    elif type(o) is Trait and o.mode == 'Fat':
        o.traits.to_excel(path + os.sep + 'traits.xlsx', index=False)
        o.trait_categories.to_excel(path + os.sep + 'trait_categories.xlsx', index=False)
        o.trait_synonyms.to_excel(path + os.sep + 'trait_synonyms.xlsx', index=False)
        o.trait_mapped_terms.to_excel(path + os.sep + 'trait_mapped_terms.xlsx', index=False)
        o.associated_pgs_ids.to_excel(path + os.sep + 'associated_pgs_ids.xlsx', index=False)
        o.child_associated_pgs_ids.to_excel(path + os.sep + 'child_associated_pgs_ids.xlsx', index=False)
    elif type(o) is TraitCategory and o.mode == 'Fat':
        o.efotraits.to_excel(path + os.sep + 'efotraits.xlsx', index=False)
        o.trait_categories.to_excel(path + os.sep + 'trait_categories.xlsx', index=False)
    else:
        raise Exception(
            "Failed to convert to EXCEL files. "
            "Please check whether the type of the input object is supported by pandasPGS, "
            "and check whether the mode attribute of the input object is 'Fat'.")


def read_scoring_file(pgs_id: str = None, grch: str = 'GRCh37') -> DataFrame:
    """
    Download a scoring file and convert it to a DataFrame. The directory of the downloaded file is $HOME/pandaspgs_home.

    Args:
        pgs_id: Polygenic Score ID.
        grch: GRCh37 or GRCh38.

    Returns:
        A DataFrame.

    ```Python
    from pandaspgs import read_scoring_file

    df = read_scoring_file(pgs_id='PGS000737')
    ```
    """
    if pgs_id is None:
        raise Exception("pgs_id can't be None.")
    raw_score_data = get_scores(pgs_id=pgs_id).raw_data
    if len(raw_score_data) != 1:
        raise Exception("Unable to find the link to download. Please check the pgs_id.")
    url = raw_score_data[0]['ftp_harmonized_scoring_files'][grch]['positions']
    match_obj = re.match('.*/(.*)', url)
    file_name = match_obj.group(1)
    os.makedirs(home_path, exist_ok=True)

    with s.get(url, timeout=60, stream=True) as r:
        online_size = r.headers.get('content-length', 0)
        local_size = 0
        if os.path.exists(home_path + os.sep + file_name):
            local_size = os.path.getsize(home_path + os.sep + file_name)
        if local_size > 0 and (int(online_size) == local_size):
            sys.stdout.write('[SKIP]: %s has been downloaded in %s\n' % (file_name, home_path))
        else:
            r.raise_for_status()
            with open(home_path + os.sep + file_name, 'wb') as f:
                i = 0
                for chunk in r.iter_content(chunk_size=1024):
                    i += 1024
                    sys.stdout.write('%s downloading: %.2f MB\r' % (file_name, i / 1024 / 1024))
                    f.write(chunk)
            sys.stdout.write('%s(%.2f MB) has been downloaded in %s\n' % (file_name, i / 1024 / 1024, home_path))

    df = read_table(home_path + os.sep + file_name, comment='#', compression='gzip')
    return df


def genotype_weighted_score(s: Series) -> DataFrame:
    genotype = [s['effect_allele'] + '/' + s['effect_allele'], s['effect_allele'] + '/' + s['other_allele'],
                s['other_allele'] + '/' + s['other_allele']]
    weighted_score = [2 * s['effect_weight'], 1 * s['effect_weight'], 0 * s['effect_weight']]
    data = {s['rsID'] + "_genotype": genotype, s['rsID'] + "_weighted_score": weighted_score}
    return DataFrame(data=data)



from pandaspgs.client import get_cohort
from pandaspgs.cohort import Cohort


def get_cohorts(cohort_symbol: str = None, cached: bool= True, mode: str = 'Fat') -> Cohort:
    """
    Get Cohort data from the server.

    Args:
        cohort_symbol: Short name of a cohort
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A Cohort object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_cohort import get_cohorts

    ch = get_cohorts(cohort_symbol='ABCFS')
    ```
    """
    if cohort_symbol is None:
        return Cohort(get_cohort('https://www.pgscatalog.org/rest/cohort/all', cached=cached), mode)
    by_id = None
    if cohort_symbol is not None:
        by_id = get_cohort('https://www.pgscatalog.org/rest/cohort/%s' % cohort_symbol, cached=cached)
    return Cohort(by_id, mode)



import pytest
from pandaspgs import *
import os
import shutil
from sqlalchemy import create_engine

home_path = os.path.expanduser('~') + os.sep + 'pandaspgs_home'


def test_read_scoring_file():
    shutil.rmtree(home_path, ignore_errors=True)
    df3 = read_scoring_file(pgs_id='PGS002256')

    df1 = read_scoring_file(pgs_id='PGS000737')
    df2 = read_scoring_file(pgs_id='PGS000737', grch='GRCh38')
    shutil.rmtree(home_path, ignore_errors=True)
    assert df1.size == 3 * 14 and df2.size == 3 * 12
    with pytest.raises(Exception):
        read_scoring_file(pgs_id='PGS000xxxx')


def test_write_to_sql():
    engine = create_engine("mysql+pymysql://{user}:{pw}@{host}/{db}"
                           .format(host="localhost", db="pgs", user="root", pw="123456"))
    scores = get_scores()
    scores.scores.to_sql('scores', engine, index=False)
    scores.samples_variants.to_sql('samples_variants', engine, index=False)
    scores.samples_variants_cohorts.to_sql('samples_variants_cohorts', engine, index=False)
    scores.trait_efo.to_sql('trait_efo', engine, index=False)
    scores.samples_training.to_sql('samples_training', engine, index=False)
    scores.samples_training_cohorts.to_sql('samples_training_cohorts', engine, index=False)
    scores.ancestry_distribution.to_sql('ancestry_distribution', engine, index=False)

    cohort = get_cohorts()
    cohort.cohorts.to_sql('cohorts', engine, index=False)
    cohort.associated_pgs_ids.to_sql('associated_pgs_ids', engine, index=False)

    pp = get_performances()
    pp.performance_metrics.to_sql('performance_metrics', engine, index=False)
    pp.samples.to_sql('samples', engine, index=False)
    pp.cohorts.to_sql('cohorts', engine, index=False)
    pp.effect_sizes.to_sql('effect_sizes', engine, index=False)
    pp.class_acc.to_sql('class_acc', engine, index=False)
    pp.othermetrics.to_sql('othermetrics', engine, index=False)

    pub = get_publications()
    pub.publications.to_sql('publications', engine, index=False)
    pub.associated_pgs_ids.to_sql('associated_pgs_ids', engine, index=False)

    rel = get_releases()
    rel.releases.to_sql('releases', engine, index=False)
    rel.released_performance_ids.to_sql('released_performance_ids', engine, index=False)
    rel.released_publication_ids.to_sql('released_publication_ids', engine, index=False)
    rel.released_score_ids.to_sql('released_score_ids', engine, index=False)
    rel.released_new_trait_ids.to_sql('released_new_trait_ids', engine, index=False)

    ss = get_sample_sets()
    ss.sample_sets.to_sql('sample_sets', engine, index=False)
    ss.samples.to_sql('samples', engine, index=False)
    ss.cohorts.to_sql('cohorts', engine, index=False)

    trait = get_traits()
    trait.traits.to_sql('traits', engine, index=False)
    trait.trait_categories.to_sql('trait_categories', engine, index=False)
    trait.trait_synonyms.to_sql('trait_synonyms', engine, index=False)
    trait.trait_mapped_terms.to_sql('trait_mapped_terms', engine, index=False)
    trait.associated_pgs_ids.to_sql('associated_pgs_ids', engine, index=False)
    trait.child_associated_pgs_ids.to_sql('child_associated_pgs_ids', engine, index=False)

    trait_c = get_trait_categories()
    trait_c.efotraits.to_sql('efotraits', engine, index=False)
    trait_c.trait_categories.to_sql('trait_categories', engine, index=False)

    ancestry = get_ancestry_categories()
    ancestry.ancestry_categories.to_sql('ancestry_categories', engine, index=False)
    ancestry.categories.to_sql('categories', engine, index=False)


def test_write_csv():
    shutil.rmtree(home_path + os.sep, ignore_errors=True)
    os.makedirs(home_path, exist_ok=True)
    ancestry = get_ancestry_categories()
    write_csv(home_path + os.sep + 'ancestry', ancestry)
    assert os.path.isfile(home_path + os.sep + 'ancestry' + os.sep + 'ancestry_categories.csv')
    assert os.path.isfile(home_path + os.sep + 'ancestry' + os.sep + 'categories.csv')
    trait_c = get_trait_categories()
    write_csv(home_path + os.sep + 'trait_c', trait_c)
    assert os.path.isfile(home_path + os.sep + 'trait_c' + os.sep + 'trait_categories.csv')
    trait = get_traits()
    write_csv(home_path + os.sep + 'trait', trait)
    assert os.path.isfile(home_path + os.sep + 'trait' + os.sep + 'traits.csv')
    ss = get_sample_sets()
    write_csv(home_path + os.sep + 'ss', ss)
    assert os.path.isfile(home_path + os.sep + 'ss' + os.sep + 'sample_sets.csv')
    rel = get_releases()
    write_csv(home_path + os.sep + 'rel', rel)
    assert os.path.isfile(home_path + os.sep + 'rel' + os.sep + 'releases.csv')
    pub = get_publications()
    write_csv(home_path + os.sep + 'pub', pub)
    assert os.path.isfile(home_path + os.sep + 'pub' + os.sep + 'publications.csv')
    pp = get_performances()
    write_csv(home_path + os.sep + 'pp', pp)
    assert os.path.isfile(home_path + os.sep + 'pp' + os.sep + 'performance_metrics.csv')
    cohort = get_cohorts()
    write_csv(home_path + os.sep + 'cohort', cohort)
    assert os.path.isfile(home_path + os.sep + 'cohort' + os.sep + 'cohorts.csv')
    scores = get_scores()
    write_csv(home_path + os.sep + 'scores', scores)
    assert os.path.isfile(home_path + os.sep + 'scores' + os.sep + 'scores.csv')


def test_write_xlsx():
    shutil.rmtree(home_path + os.sep, ignore_errors=True)
    os.makedirs(home_path, exist_ok=True)
    ancestry = get_ancestry_categories()
    write_xlsx(home_path + os.sep + 'ancestry', ancestry)
    assert os.path.isfile(home_path + os.sep + 'ancestry' + os.sep + 'ancestry_categories.xlsx')
    assert os.path.isfile(home_path + os.sep + 'ancestry' + os.sep + 'categories.xlsx')
    trait_c = get_trait_categories()
    write_xlsx(home_path + os.sep + 'trait_c', trait_c)
    assert os.path.isfile(home_path + os.sep + 'trait_c' + os.sep + 'trait_categories.xlsx')
    trait = get_traits()
    write_xlsx(home_path + os.sep + 'trait', trait)
    assert os.path.isfile(home_path + os.sep + 'trait' + os.sep + 'traits.xlsx')
    ss = get_sample_sets()
    write_xlsx(home_path + os.sep + 'ss', ss)
    assert os.path.isfile(home_path + os.sep + 'ss' + os.sep + 'sample_sets.xlsx')
    rel = get_releases()
    write_xlsx(home_path + os.sep + 'rel', rel)
    assert os.path.isfile(home_path + os.sep + 'rel' + os.sep + 'releases.xlsx')
    pub = get_publications()
    write_xlsx(home_path + os.sep + 'pub', pub)
    assert os.path.isfile(home_path + os.sep + 'pub' + os.sep + 'publications.xlsx')
    pp = get_performances()
    write_xlsx(home_path + os.sep + 'pp', pp)
    assert os.path.isfile(home_path + os.sep + 'pp' + os.sep + 'performance_metrics.xlsx')
    cohort = get_cohorts()
    write_xlsx(home_path + os.sep + 'cohort', cohort)
    assert os.path.isfile(home_path + os.sep + 'cohort' + os.sep + 'cohorts.xlsx')
    scores = get_scores()
    write_xlsx(home_path + os.sep + 'scores', scores)
    assert os.path.isfile(home_path + os.sep + 'scores' + os.sep + 'scores.xlsx')



from pandaspgs.client import get_data, get_publication, clear_cache, get_ancestry_category
from cachetools import TTLCache

cache = TTLCache(maxsize=1024, ttl=60)


def test_get_data():

    r1 = get_data('https://www.pgscatalog.org/rest/publication/search?pgs_id=PGS000001', cache_impl=cache)
    assert len(r1) == 1
    r2 = get_data('https://www.pgscatalog.org/rest/publication/PGP000001', cache_impl=cache)
    assert len(r2) == 1
    r3 = get_data('https://www.pgscatalog.org/rest/publication/all', cache_impl=cache)
    r3 = get_data('https://www.pgscatalog.org/rest/publication/all', cache_impl=cache)
    assert len(r3) == 455
    r4 = get_data('https://www.pgscatalog.org/rest/publication/all', cache_impl=cache, cached=False)
    assert len(r3) == 455


def test_get_publication():
    r1 = get_publication('https://www.pgscatalog.org/rest/publication/search?pgs_id=PGS000001')
    assert len(r1) == 1
    r2 = get_publication('https://www.pgscatalog.org/rest/publication/PGP000001')
    assert len(r2) == 1
    r3 = get_publication('https://www.pgscatalog.org/rest/publication/all')
    r3 = get_publication('https://www.pgscatalog.org/rest/publication/all')
    clear_cache('Publication')
    r3 = get_publication('https://www.pgscatalog.org/rest/publication/all')
    assert len(r3) == 455
    r4 = get_publication('https://www.pgscatalog.org/rest/publication/all', cached=False)
    assert len(r3) == 455


def test_get_ancestry_category():
    cat = get_ancestry_category('https://www.pgscatalog.org/rest/ancestry_categories')
    assert cat[0]['symbol'] == 'AFR'



from pandaspgs.client import get_publication
from pandaspgs.publication import Publication


def get_publications(pgs_id: str = None, pgp_id: str = None, pmid: int = None, author: str = None, cached: bool = True,
                     mode: str = 'Fat') -> Publication:
    """
    Get Publication data from the server.

    Args:
        pgs_id: PGS Publication ID (PGP)
        pgp_id: Polygenic Score ID (PGS)
        pmid: PubMed ID (without the prefix "PMID:")
        author: Publication author (any author in the list of authors in a publication)
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A Publication object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_publication import get_publications

    ch = get_publications(pgp_id='PGP000003')
    ```
    """
    if pgs_id is None and pgp_id is None and pmid is None and author is None:
        return Publication(get_publication('https://www.pgscatalog.org/rest/publication/all', cached=cached), mode)
    by_id = None
    by_other = None
    if pgp_id is not None:
        by_id = get_publication('https://www.pgscatalog.org/rest/publication/%s' % pgp_id, cached=cached)
    if pgs_id is not None or pmid is not None or author is not None:
        query_str = []
        if pgs_id is not None:
            query_str.append('pgs_id=%s' % pgs_id)
        if pmid is not None:
            query_str.append('pmid=%d' % pmid)
        if author is not None:
            query_str.append('author=%s' % author)
        by_other = get_publication('https://www.pgscatalog.org/rest/publication/search?%s' % '&'.join(query_str))
    if pgp_id is None:
        return Publication(by_other, mode)
    if pgs_id is None and pmid is None and author is None:
        return Publication(by_id, mode)
    other_set = set()
    id_dict = {}
    for single in by_id:
        id_dict[single['id']] = single
    pgp_id_set = id_dict.keys()
    for single in by_other:
        other_set.add(single['id'])
    intersection = pgp_id_set & other_set
    result = []
    for id in intersection:
        result.append(id_dict[id])
    return Publication(result, mode)









from pandas import DataFrame, json_normalize, set_option

set_option('display.max_columns', None)
set_option('display.width', 1000)
set_option('display.colheader_justify', 'center')
set_option('display.precision', 3)


class Trait:
    """
       An object that stores data of type Trait. To understand the significance of each column of the DataFrame. Please visit "EFOTrait_Ontology" in [PGS Catalog Documentation](https://www.pgscatalog.org/rest/) for details.

       Attributes:
            raw_data: list. Convert from obtained JSON data
            traits : DataFrame. It only exists if the parameter mode of constructor is Fat.
            trait_categories: DataFrame. It only exists if the parameter mode of constructor is Fat.
            trait_synonyms: DataFrame. It only exists if the parameter mode of constructor is Fat.
            trait_mapped_terms: DataFrame. It only exists if the parameter mode of constructor is Fat.
            associated_pgs_ids: DataFrame. It only exists if the parameter mode of constructor is Fat.
            child_associated_pgs_ids: DataFrame. It only exists if the parameter mode of constructor is Fat.
            mode: Fat or Thin. Specifies the mode of the returned object.

       ```python
       from pandaspgs.get_trait import get_traits

       ch = get_traits(trait_id='EFO_0004214')
       ch
       ch.raw_data
       ch.mode
       ch.traits
       ch.trait_categories
       ch.trait_synonyms
       ch.trait_mapped_terms
       ch.associated_pgs_ids
       ch.child_associated_pgs_ids
       ```
       Subset object s by either identifier or position
       ```python
       all_df = get_traits()
       all_df[0].traits
       all_df[0:3].traits
       all_df['EFO_0004214'].traits
       all_df[('EFO_0004214','HP_0002027','HP_0011458')].traits
       ```
       Objects can be manipulated like sets in the mathematical sense.
       ```python
       one = get_scores(trait_id='EFO_0004214')
       two = get_scores(trait_id='HP_0002027')
       three = get_scores(trait_id='HP_0011458')
       one_and_two = one+two
       two_and_three = two+three
       only_one = one_and_two - two_and_three
       only_two = one_and_two & two_and_three
       one_and_two_and_three = one_and_two | two_and_three
       one_and_three = one_and_two ^ two_and_three
       ```
       """
    def __init__(self, data: list = [], mode: str = "Fat"):
        """
        Args:
            data: Raw JSON data.
            mode: Fat or Thin. Specifies the mode of the object.
        """
        if data is None:
            data = []
        if mode not in ['Thin', "Fat"]:
            raise Exception("Mode must be Fat or Thin")
        self.raw_data = data
        self.mode = mode
        if mode == "Thin":
            return
        if data is None or len(data) == 0:
            self.traits = DataFrame(
                columns=['id', 'label', 'description', 'url'])
            self.trait_categories = DataFrame(
                columns=['trait _id', 'trait_category'])
            self.trait_synonyms = DataFrame(columns=['trait_id', 'trait_synonym'])
            self.trait_mapped_terms = DataFrame(columns=['trait_id', 'trait_mapped_term'])
            self.associated_pgs_ids = DataFrame(columns=['trait_id', 'associated_pgs_id'])
            self.child_associated_pgs_ids = DataFrame(columns=['trait_id', 'child_associated_pgs_id'])
            return
        datas = json_normalize(data=data, max_level=1).drop(columns=['id', 'label', 'description', 'url'])
        datas['trait_categories'] = datas['trait_categories'].map(lambda x: x == [])
        datas['trait_synonyms'] = datas['trait_synonyms'].map(lambda x: x == [])
        datas['trait_mapped_terms'] = datas['trait_mapped_terms'].map(lambda x: x == [])
        datas['associated_pgs_ids'] = datas['associated_pgs_ids'].map(lambda x: x == [])
        datas['child_associated_pgs_ids'] = datas['child_associated_pgs_ids'].map(lambda x: x == [])
        self.traits = json_normalize(data=data, max_level=1).drop(
            columns=['trait_categories', 'trait_synonyms', 'trait_mapped_terms',
                     'associated_pgs_ids', 'child_associated_pgs_ids'])
        if not datas['trait_categories'].all():
            self.trait_categories = json_normalize(data=data, record_path=['trait_categories'], meta=['id'])
            self.trait_categories.columns = ['trait_category', 'trait_id']
        else:
            self.trait_categories = DataFrame(
                columns=['trait _id', 'trait_category'])
        if not datas['trait_synonyms'].all():
            self.trait_synonyms = json_normalize(data=data, record_path=['trait_synonyms'], meta=['id'])
            self.trait_synonyms.columns = ['trait_synonym', 'trait_id']
        else:
            self.trait_synonyms = DataFrame(columns=['trait_id', 'trait_synonym'])
        if not datas['trait_mapped_terms'].all():
            self.trait_mapped_terms = json_normalize(data=data, record_path=['trait_mapped_terms'], meta=['id'])
            self.trait_mapped_terms.columns = ['trait_mapped_term', 'trait_id']
        else:
            self.trait_mapped_terms = DataFrame(columns=['trait_id', 'trait_mapped_term'])
        if not datas['associated_pgs_ids'].all():
            self.associated_pgs_ids = json_normalize(data=data, record_path=['associated_pgs_ids'], meta=['id'])
            self.associated_pgs_ids.columns = ['associated_pgs_id', 'trait_id']
        else:
            self.associated_pgs_ids = DataFrame(columns=['trait_id', 'associated_pgs_id'])
        if not datas['child_associated_pgs_ids'].all():
            self.child_associated_pgs_ids = json_normalize(data=data, record_path=['child_associated_pgs_ids'],
                                                           meta=['id'])
            self.child_associated_pgs_ids.columns = ['child_associated_pgs_id', 'trait_id']
        else:
            self.child_associated_pgs_ids = DataFrame(columns=['trait_id', 'child_associated_pgs_id'])
        return

    def __str__(self):
        if self.mode == 'Fat':
            return ("Trait is running in fat mode. It has 6 DataFrames with hierarchical dependencies.\ntraits: "
                    "%d rows\n|\n -associated_pgs_ids: %d rows\n|\n -child_associated_pgs_ids:"
                    "%d rows\n|\n -trait_categories: %d rows\n|\n -trait_mapped_terms: %d rows\n|\n -trait_synonyms:"
                    " %d rows" % (
                        len(self.traits), len(self.associated_pgs_ids), len(self.child_associated_pgs_ids),
                        len(self.trait_categories), len(self.trait_mapped_terms), len(self.trait_synonyms)))
        if self.mode == 'Thin':
            return ('Trait is running in thin mode. It has 1 list that contains the raw data.\nraw_data: a list of '
                    'size x.')

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        if isinstance(item, str) or isinstance(item, int):
            arr = [item]
        elif isinstance(item, list) or isinstance(item, tuple) or isinstance(item, range):
            arr = item
        elif isinstance(item, slice):
            start, stop, step = item.indices(len(self))
            arr = list(range(start, stop, step))
        else:
            raise TypeError('Invalid argument type：{}'.format(type(item)))
        raw_data = self.raw_data
        raw_data_dict = {}
        for j in raw_data:
            raw_data_dict[j['id']] = j
        sub_set = []
        for i in arr:
            if isinstance(i, str):
                sub_set.append(raw_data_dict[i])
            elif isinstance(i, int):
                sub_set.append(raw_data[i])
            else:
                raise TypeError('Invalid item type: {}'.format(type(i)))
        return Trait(sub_set, self.mode)

    def __add__(self, other):
        if self.mode == other.mode:
            return Trait(self.raw_data + other.raw_data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __sub__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set - other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Trait(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __and__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set & other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Trait(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __or__(self, other):
        if self.mode == other.mode:
            and_dict = {}
            for i in self.raw_data:
                and_dict[i['id']] = i
            for j in other.raw_data:
                and_dict[j['id']] = j
            data = list(and_dict.values())
            return Trait(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __xor__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            and_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                and_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
                and_dict[j['id']] = j
            sub_key = self_key_set ^ other_key_set
            data = []
            for k in sub_key:
                data.append(and_dict[k])
            return Trait(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __eq__(self, other):
        if self is None or other is None:
            return self is None and other is None
        return self.raw_data == other.raw_data and self.mode == other.mode

    def __len__(self):
        return len(self.raw_data)



from pandaspgs.client import get_trait_category, get_trait
from pandaspgs.trait import Trait
from pandaspgs.trait_category import TraitCategory


def get_trait_categories(cached: bool = True, mode: str = 'Fat') -> TraitCategory:
    """
    Get TraitCategory data from the server.

    Args:
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A TraitCategory object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_trait_category import get_trait_categories

    ch = get_trait_categories()
    ```
    """
    return TraitCategory(get_trait_category('https://www.pgscatalog.org/rest/trait_category/all', cached), mode)


def get_traits(trait_id: str = None, term: str = None, exact: bool = None, cached: bool = True, mode: str = 'Fat') -> Trait:
    """
    Get Trait data from the server.

    Args:
        trait_id: Ontology ID (e.g. from EFO, HP or MONDO) with the format "EFO_XXXX"
        term: Trait ID, trait name, trait category, trait synonym or external mapped terms/IDs
        exact: Flag to search only the exact term or not
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A Trait object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_trait import get_traits

    ch = get_traits(trait_id='EFO_0000305')
    ```
    """
    if exact is not None:
        if exact:
            num = "1"
        else:
            num = "0"

    if trait_id is None and term is None and exact is None:
        return Trait(
            get_trait('https://www.pgscatalog.org/rest/trait/all?include_child_associated_pgs_ids=1', cached=cached),
            mode)
    elif term is None and exact is not None:
        raise Exception("exact is available only if term is not None")
    elif trait_id is not None:
        by_other = get_trait('https://www.pgscatalog.org/rest/trait/%s?include_children=0' % trait_id, cached=cached)
        if term is None:
            return Trait(by_other, mode)
        else:
            if exact is not None:
                by_pgs_id = get_trait(
                    "https://www.pgscatalog.org/rest/trait/search?include_children=0&term=%s&exact=%s" % (term, num),
                    cached=cached)
            else:
                by_pgs_id = get_trait("https://www.pgscatalog.org/rest/trait/search?include_children=0&term=%s" % term,
                                      cached=cached)
        other_set = set()
        pgs_id_dict = {}
        for single in by_pgs_id:
            pgs_id_dict[single['id']] = single
        pgs_id_set = pgs_id_dict.keys()
        for single in by_other:
            other_set.add(single['id'])
        intersection = pgs_id_set & other_set
        result = []
        for id in intersection:
            result.append(pgs_id_dict[id])
        return Trait(result, mode)
    else:
        if exact is not None:
            return Trait(get_trait(
                "https://www.pgscatalog.org/rest/trait/search?include_children=0&term=%s&exact=%s" % (term, num),
                cached=cached), mode)
        else:
            return Trait(get_trait("https://www.pgscatalog.org/rest/trait/search?include_children=0&term=%s" % term,
                                   cached=cached), mode)


def get_child_traits(trait_id: str = None, cached: bool =True, mode: str = 'Fat') -> Trait:
    """
    Get children Trait data of the specified Trait ID from the server.

    Args:
        trait_id: Ontology ID (e.g. from EFO, HP or MONDO) with the format "EFO_XXXX"
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A Trait object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_child_trait import get_child_traits

    ch = get_child_traits(trait_id='EFO_0000305')
    ```
    """
    return Trait(get_trait('https://www.pgscatalog.org/rest/trait/%s?include_children=1' % trait_id, cached=cached),
                 mode)






from pandas import DataFrame, json_normalize, set_option

set_option('display.max_columns', None)
set_option('display.width', 1000)
set_option('display.colheader_justify', 'center')
set_option('display.precision', 3)


class Release:
    """
       An object that stores data of type Release. To understand the significance of each column of the DataFrame. Please visit "Release" in [PGS Catalog Documentation](https://www.pgscatalog.org/rest/) for details.

       Attributes:
            raw_data: list. Convert from obtained JSON data
            releases: DataFrame. It only exists if the parameter mode of constructor is Fat.
            released_performance_ids: DataFrame. It only exists if the parameter mode of constructor is Fat.
            released_publication_ids: DataFrame. It only exists if the parameter mode of constructor is Fat.
            released_score_ids: DataFrame. It only exists if the parameter mode of constructor is Fat.
            released_new_trait_ids: DataFrame. It only exists if the parameter mode of constructor is Fat.
            mode: Fat or Thin. Specifies the mode of the returned object.

       ```python
       from pandaspgs.get_release import get_releases

       ch = get_releases(date="2024-01-26")
       ch
       ch.raw_data
       ch.mode
       ch.releases
       ch.released_performance_ids
       ch.released_publication_ids
       ch.released_score_ids
       ch.released_new_trait_ids
       ```
       Subset object s by either identifier or position
       ```python
       all_df = get_releases()
       all_df[0].releases
       all_df[0:3].releases
       all_df['2024-01-26'].releases
       all_df[('2024-01-26','2024-01-11','2023-12-19')].releases
       ```
       Objects can be manipulated like sets in the mathematical sense.
       ```python
       one = get_releases(date='2024-01-26')
       two = get_releases(date='2024-01-11')
       three = get_releases(date='2023-12-19')
       one_and_two = one+two
       two_and_three = two+three
       only_one = one_and_two - two_and_three
       only_two = one_and_two & two_and_three
       one_and_two_and_three = one_and_two | two_and_three
       one_and_three = one_and_two ^ two_and_three
       ```
       """
    def __init__(self, data: list = [], mode: str = "Fat"):
        """
        Args:
            data: Raw JSON data.
            mode: Fat or Thin. Specifies the mode of the object.
        """
        if data is None:
            data = []
        if mode not in ['Thin', "Fat"]:
            raise Exception("Mode must be Fat or Thin")
        self.raw_data = data
        self.mode = mode
        if mode == "Thin":
            return
        if data is None or len(data) == 0:
            self.releases = DataFrame(columns=['date', 'score_count', 'performance_count', 'publication_count', 'notes'])
            self.released_score_ids = DataFrame(columns=['release_date', 'released_score_id'])
            self.released_publication_ids = DataFrame(columns=['release_date', 'released_publication'])
            self.released_performance_ids = DataFrame(columns=['release_date', 'released_performance_id'])
            self.released_new_trait_ids = DataFrame(columns=['release_date','released_new_trait_id'])
            return
        datas = json_normalize(data=data, max_level=1)
        datas['released_score_ids'] = datas['released_score_ids'].map(lambda x: x == [])
        datas['released_publication_ids'] = datas['released_publication_ids'].map(lambda x: x == [])
        datas['released_performance_ids'] = datas['released_performance_ids'].map(lambda x: x == [])
        datas['released_new_trait_ids'] = datas['released_new_trait_ids'].map(lambda x: x == [])
        self.releases = json_normalize(data=data, max_level=1).drop(
            columns=['released_score_ids', 'released_publication_ids', 'released_performance_ids', 'released_new_trait_ids'])
        if not datas['released_score_ids'].all():
            self.released_score_ids = json_normalize(data=data, record_path=['released_score_ids'], meta=['date'])
            self.released_score_ids.columns = ['released_score_id', 'release_date']
        else:
            self.released_score_ids = DataFrame(
                columns=['released_score_id', 'release_date'])
        if not datas['released_publication_ids'].all():
            self.released_publication_ids = json_normalize(data=data, record_path=['released_publication_ids'],
                                                           meta=['date'])
            self.released_publication_ids.columns = ['released_publication_id', 'release_date']
        else:
            self.released_publication_ids = DataFrame(columns=['released_publication_id', 'release_date'])
        if not datas['released_performance_ids'].all():
            self.released_performance_ids = json_normalize(data=data, record_path=['released_performance_ids'],
                                                           meta=['date'])
            self.released_performance_ids.columns = ['released_performance_id', 'release_date']
        else:
            self.released_performance_ids = DataFrame(columns=['released_performance_id', 'release_date'])
        if not datas['released_new_trait_ids'].all():
            self.released_new_trait_ids = json_normalize(data=data, record_path=['released_new_trait_ids'],
                                                           meta=['date'])
            self.released_new_trait_ids.columns = ['released_new_trait_ids', 'release_date']
        else:
            self.released_new_trait_ids = DataFrame(columns=['released_new_trait_ids', 'release_date'])
        return

    def __str__(self):
        if self.mode == 'Fat':
            return ("Release is running in fat mode. It has 4 DataFrames with hierarchical dependencies.\n-releases: "
                    "%d rows\n|\n -released_score_ids: %d rows\n|\n -released_publication_ids: "
                    "%d rows\n|\n -released_performance_ids: %d rows" % (
                        len(self.releases), len(self.released_score_ids), len(self.released_publication_ids),
                        len(self.released_performance_ids)))
        if self.mode == 'Thin':
            return ('Release is running in thin mode. It has 1 list that contains the raw data.\nraw_data: a list of '
                    'size %d.') % len(self.raw_data)

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        if isinstance(item, str) or isinstance(item, int):
            arr = [item]
        elif isinstance(item, list) or isinstance(item, tuple) or isinstance(item, range):
            arr = item
        elif isinstance(item, slice):
            start, stop, step = item.indices(len(self))
            arr = list(range(start, stop, step))
        else:
            raise TypeError('Invalid argument type：{}'.format(type(item)))
        raw_data = self.raw_data
        raw_data_dict = {}
        for j in raw_data:
            raw_data_dict[j['date']] = j
        sub_set = []
        for i in arr:
            if isinstance(i, str):
                sub_set.append(raw_data_dict[i])
            elif isinstance(i, int):
                sub_set.append(raw_data[i])
            else:
                raise TypeError('Invalid item type: {}'.format(type(i)))
        return Release(sub_set, self.mode)

    def __add__(self, other):
        if self.mode == other.mode:
            return Release(self.raw_data + other.raw_data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __sub__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['date'])
                self_dict[i['date']] = i
            for j in other.raw_data:
                other_key_set.add(j['date'])
            sub_key = self_key_set - other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Release(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __and__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['date'])
                self_dict[i['date']] = i
            for j in other.raw_data:
                other_key_set.add(j['date'])
            sub_key = self_key_set & other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Release(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __or__(self, other):
        if self.mode == other.mode:
            and_dict = {}
            for i in self.raw_data:
                and_dict[i['date']] = i
            for j in other.raw_data:
                and_dict[j['date']] = j
            data = list(and_dict.values())
            return Release(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __xor__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            and_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['date'])
                and_dict[i['date']] = i
            for j in other.raw_data:
                other_key_set.add(j['date'])
                and_dict[j['date']] = j
            sub_key = self_key_set ^ other_key_set
            data = []
            for k in sub_key:
                data.append(and_dict[k])
            return Release(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __eq__(self, other):
        if self is None or other is None:
            return self is None and other is None
        return self.raw_data == other.raw_data and self.mode == other.mode

    def __len__(self):
        return len(self.raw_data)



from pandas import DataFrame, json_normalize, set_option, Series

set_option('display.max_columns', None)
set_option('display.width', 1000)
set_option('display.colheader_justify', 'center')
set_option('display.precision', 3)
import numpy


class Score:
    """
       An object that stores data of type Score. To understand the significance of each column of the DataFrame. Please visit "Score" in [PGS Catalog Documentation](https://www.pgscatalog.org/rest/) for details.

       Attributes:
            raw_data: list. Convert from obtained JSON data
            scores: DataFrame. It only exists if the parameter mode of constructor is Fat.
            samples_variants: DataFrame. It only exists if the parameter mode of constructor is Fat.
            samples_variants_cohorts: DataFrame. It only exists if the parameter mode of constructor is Fat.
            trait_efo: DataFrame. It only exists if the parameter mode of constructor is Fat.
            samples_training: DataFrame. It only exists if the parameter mode of constructor is Fat.
            samples_training_cohorts: DataFrame. It only exists if the parameter mode of constructor is Fat.
            ancestry_distribution: DataFrame. It only exists if the parameter mode of constructor is Fat.
            mode: Fat or Thin. Specifies the mode of the returned object.

       ```python
       from pandaspgs.get_score import get_scores

       ch = get_scores(pgs_id='PGS000001')
       ch
       ch.raw_data
       ch.mode
       ch.scores
       ch.samples_variants
       ch.samples_variants_cohorts
       ch.trait_efo
       ch.samples_training
       ch.samples_training_cohorts
       ch.ancestry_distribution
       ```
       Subset object s by either identifier or position
       ```python
       all_df = get_scores()
       all_df[0].scores
       all_df[0:3].scores
       all_df['PGS000001'].scores
       all_df[('PGS000001','PGS000002','PGS000003')].scores
       ```
       Objects can be manipulated like sets in the mathematical sense.
       ```python
       one = get_scores(pgs_id='PGS000001')
       two = get_scores(pgs_id='PGS000002')
       three = get_scores(pgs_id='PGS000003')
       one_and_two = one+two
       two_and_three = two+three
       only_one = one_and_two - two_and_three
       only_two = one_and_two & two_and_three
       one_and_two_and_three = one_and_two | two_and_three
       one_and_three = one_and_two ^ two_and_three
       ```
       """

    def __init__(self, data: list = [], mode: str = "Fat"):
        """
        Args:
            data: Raw JSON data.
            mode: Fat or Thin. Specifies the mode of the object.
        """
        if data is None:
            data = []
        if mode not in ['Thin', "Fat"]:
            raise Exception("Mode must be Fat or Thin")
        self.raw_data = data
        self.mode = mode
        if mode == "Thin":
            return
        if data is None or len(data) == 0:
            self.scores = DataFrame(
                columns=['id'
                    , 'name'
                    , 'ftp_scoring_file'
                    , 'ftp_harmonized_scoring_files.GRCh37.positions'
                    , 'ftp_harmonized_scoring_files.GRCh38.positions'
                    , 'publication.id'
                    , 'publication.title'
                    , 'publication.doi'
                    , 'publication.PMID'
                    , 'publication.journal'
                    , 'publication.firstauthor'
                    , 'publication.date_publication'
                    , 'matches_publication'
                    , 'trait_reported'
                    , 'trait_additional'
                    , 'method_name'
                    , 'method_params'
                    , 'variants_number'
                    , 'variants_interactions'
                    , 'variants_genomebuild'
                    , 'weight_type'
                    , 'date_release'
                    , 'license'
                         ])
            self.samples_variants = DataFrame(
                columns=['id'
                    , 'score_id'
                    , 'sample_number'
                    , 'sample_cases'
                    , 'sample_controls'
                    , 'sample_percent_male'
                    , 'sample_age.estimate_type'
                    , 'sample_age.estimate'
                    , 'sample_age.interval.type'
                    , 'sample_age.interval.lower'
                    , 'sample_age.interval.upper'
                    , 'sample_age.variability_type'
                    , 'sample_age.variability'
                    , 'sample_age.unit'
                    , 'phenotyping_free'
                    , 'followup_time.estimate_type'
                    , 'followup_time.estimate'
                    , 'followup_time.interval.type'
                    , 'followup_time.interval.lower'
                    , 'followup_time.interval.upper'
                    , 'followup_time.variability_type'
                    , 'followup_time.variability'
                    , 'followup_time.unit'
                    , 'ancestry_broad'
                    , 'ancestry_free'
                    , 'ancestry_country'
                    , 'ancestry_additional'
                    , 'source_GWAS_catalog'
                    , 'source_PMID'
                    , 'source_DOI'
                    , 'cohorts_additional'
                         ])
            self.samples_variants_cohorts = DataFrame(
                columns=['score_id'
                    , 'sample_id'
                    , 'name_short'
                    , 'name_full'
                    , 'name_others'])
            self.trait_efo = DataFrame(
                columns=['score_id'
                    , 'id'
                    , 'label'
                    , 'description'
                    , 'url'
                         ])
            self.samples_training = DataFrame(
                columns=['id'
                    , 'score_id'
                    , 'sample_number'
                    , 'sample_cases'
                    , 'sample_controls'
                    , 'sample_percent_male'
                    , 'sample_age.estimate_type'
                    , 'sample_age.estimate'
                    , 'sample_age.interval.type'
                    , 'sample_age.interval.lower'
                    , 'sample_age.interval.upper'
                    , 'sample_age.variability_type'
                    , 'sample_age.variability'
                    , 'sample_age.unit'
                    , 'phenotyping_free'
                    , 'followup_time.estimate_type'
                    , 'followup_time.estimate'
                    , 'followup_time.interval.type'
                    , 'followup_time.interval.lower'
                    , 'followup_time.interval.upper'
                    , 'followup_time.variability_type'
                    , 'followup_time.variability'
                    , 'followup_time.unit'
                    , 'ancestry_broad'
                    , 'ancestry_free'
                    , 'ancestry_country'
                    , 'ancestry_additional'
                    , 'source_GWAS_catalog'
                    , 'source_PMID'
                    , 'source_DOI'
                    , 'cohorts_additional'
                         ])
            self.samples_training_cohorts = DataFrame(
                columns=['score_id'
                    , 'sample_id'
                    , 'name_short'
                    , 'name_full'
                    , 'name_others'
                         ])
            self.ancestry_distribution = DataFrame(
                columns=['score_id'
                    , 'stage'
                    , 'dist'
                    , 'count'
                    , 'multi'])
            return
        datas = json_normalize(data=data, max_level=1)
        datas['samples_variants'] = datas['samples_variants'].map(lambda x: x == [])
        datas['samples_training'] = datas['samples_training'].map(lambda x: x == [])
        datas['trait_efo'] = datas['trait_efo'].map(lambda x: x == [])

        self.scores = json_normalize(data=data, max_level=1).drop(
            columns=['samples_variants', 'samples_training',
                     'trait_efo'])
        ancestry_distribution_to_drop = ['ancestry_distribution.eval', 'ancestry_distribution.gwas',
                                         'ancestry_distribution.dev']
        for to_drop in ancestry_distribution_to_drop:
            if to_drop in self.scores.columns:
                self.scores = self.scores.drop(columns=[to_drop])
        self.scores['ftp_harmonized_scoring_files.GRCh38.positions'] = self.scores[
            'ftp_harmonized_scoring_files.GRCh38'].map(
            lambda x: x['positions'])
        self.scores['ftp_harmonized_scoring_files.GRCh37.positions'] = self.scores[
            'ftp_harmonized_scoring_files.GRCh37'].map(
            lambda x: x['positions'])
        self.scores = self.scores.drop(
            columns=['ftp_harmonized_scoring_files.GRCh38', 'ftp_harmonized_scoring_files.GRCh37'])
        if not datas['samples_variants'].all():
            self.samples_variants = json_normalize(data=data, record_path=['samples_variants'], meta=['id'])
            self.samples_variants['score_id'] = self.samples_variants['id']
            self.samples_variants['id'] = Series(data=range(0, len(self.samples_variants)))
            cohort = self.samples_variants[['id', 'score_id', 'cohorts']].copy()
            self.samples_variants = self.samples_variants.drop(columns=['cohorts'])
            cohort['sample_id'] = cohort['id']
            cohort['cohorts'] = cohort['cohorts'].apply(lambda x: x if len(x) > 0 else numpy.nan)
            cohort = cohort.dropna()
            cohort = cohort.explode('cohorts')
            if len(cohort) == 0:
                self.samples_variants_cohorts = DataFrame(
                    columns=['score_id'
                        , 'sample_id'
                        , 'name_short'
                        , 'name_full'
                        , 'name_others'])
            else:
                cohort[['name_short', 'name_full', 'name_others']] = cohort['cohorts'].apply(
                    lambda x: Series(data=[x['name_short'], x['name_full'], x['name_others']]))
                cohort = cohort.drop(columns=['id', 'cohorts'])
                self.samples_variants_cohorts = cohort

        else:
            self.samples_variants = DataFrame(
                columns=['id'
                    , 'score_id'
                    , 'sample_number'
                    , 'sample_cases'
                    , 'sample_controls'
                    , 'sample_percent_male'
                    , 'sample_age.estimate_type'
                    , 'sample_age.estimate'
                    , 'sample_age.interval.type'
                    , 'sample_age.interval.lower'
                    , 'sample_age.interval.upper'
                    , 'sample_age.variability_type'
                    , 'sample_age.variability'
                    , 'sample_age.unit'
                    , 'phenotyping_free'
                    , 'followup_time.estimate_type'
                    , 'followup_time.estimate'
                    , 'followup_time.interval.type'
                    , 'followup_time.interval.lower'
                    , 'followup_time.interval.upper'
                    , 'followup_time.variability_type'
                    , 'followup_time.variability'
                    , 'followup_time.unit'
                    , 'ancestry_broad'
                    , 'ancestry_free'
                    , 'ancestry_country'
                    , 'ancestry_additional'
                    , 'source_GWAS_catalog'
                    , 'source_PMID'
                    , 'source_DOI'
                    , 'cohorts_additional'
                         ])
            self.samples_variants_cohorts = DataFrame(
                columns=['score_id'
                    , 'sample_id'
                    , 'name_short'
                    , 'name_full'
                    , 'name_others'])
        if not datas['samples_training'].all():
            self.samples_training = json_normalize(data=data, record_path=['samples_training'], meta=['id'])
            self.samples_training['score_id'] = self.samples_training['id']
            self.samples_training['id'] = Series(data=range(0, len(self.samples_training)))
            cohort = self.samples_training[['id', 'score_id', 'cohorts']].copy()
            self.samples_training = self.samples_training.drop(columns=['cohorts'])
            cohort['sample_id'] = cohort['id']
            cohort['cohorts'] = cohort['cohorts'].apply(lambda x: x if len(x) > 0 else numpy.nan)
            cohort = cohort.dropna()
            cohort = cohort.explode('cohorts')
            if len(cohort) == 0:
                self.samples_training_cohorts = DataFrame(
                    columns=['score_id'
                        , 'sample_id'
                        , 'name_short'
                        , 'name_full'
                        , 'name_others'])
            else:
                cohort[['name_short', 'name_full', 'name_others']] = cohort['cohorts'].apply(
                    lambda x: Series(data=[x['name_short'], x['name_full'], x['name_others']]))
                cohort = cohort.drop(columns=['id', 'cohorts'])
                self.samples_training_cohorts = cohort

        else:
            self.samples_training = DataFrame(
                columns=['id'
                    , 'score_id'
                    , 'sample_number'
                    , 'sample_cases'
                    , 'sample_controls'
                    , 'sample_percent_male'
                    , 'sample_age.estimate_type'
                    , 'sample_age.estimate'
                    , 'sample_age.interval.type'
                    , 'sample_age.interval.lower'
                    , 'sample_age.interval.upper'
                    , 'sample_age.variability_type'
                    , 'sample_age.variability'
                    , 'sample_age.unit'
                    , 'phenotyping_free'
                    , 'followup_time.estimate_type'
                    , 'followup_time.estimate'
                    , 'followup_time.interval.type'
                    , 'followup_time.interval.lower'
                    , 'followup_time.interval.upper'
                    , 'followup_time.variability_type'
                    , 'followup_time.variability'
                    , 'followup_time.unit'
                    , 'ancestry_broad'
                    , 'ancestry_free'
                    , 'ancestry_country'
                    , 'ancestry_additional'
                    , 'source_GWAS_catalog'
                    , 'source_PMID'
                    , 'source_DOI'
                    , 'cohorts_additional'
                         ])
            self.samples_training_cohorts = DataFrame(
                columns=['score_id'
                    , 'sample_id'
                    , 'name_short'
                    , 'name_full'
                    , 'name_others'
                         ])
        if not datas['trait_efo'].all():
            self.trait_efo = json_normalize(data=data, record_path=['trait_efo'], meta='id', meta_prefix='score_')

        else:
            self.trait_efo = DataFrame(
                columns=['score_id'
                    , 'id'
                    , 'label'
                    , 'description'
                    , 'url'])
        raw_ancestry_distribution = json_normalize(data=data, max_level=0)
        raw_ancestry_distribution['ancestry_distribution'] = raw_ancestry_distribution['ancestry_distribution'].map(
            lambda x: [{y: x[y]} for y in x])
        raw_ancestry_distribution = raw_ancestry_distribution.explode('ancestry_distribution')
        raw_ancestry_distribution['stage'] = raw_ancestry_distribution['ancestry_distribution'].map(
            lambda x: list(x.keys())[0])
        raw_ancestry_distribution['ancestry_distribution'] = raw_ancestry_distribution['ancestry_distribution'].map(
            lambda x: list(x.values())[0])
        raw_ancestry_distribution['dist'] = raw_ancestry_distribution['ancestry_distribution'].map(
            lambda x: x['dist'])
        raw_ancestry_distribution['count'] = raw_ancestry_distribution['ancestry_distribution'].map(
            lambda x: x['count'])
        raw_ancestry_distribution['multi'] = raw_ancestry_distribution['ancestry_distribution'].map(
            lambda x: x['multi'] if 'multi' in x else None)
        raw_ancestry_distribution['score_id'] = raw_ancestry_distribution['id']
        self.ancestry_distribution = raw_ancestry_distribution[
            ['score_id', 'stage', 'dist', 'count', 'multi']].copy().reset_index(drop=True)

        if 'publication' in self.scores.columns:
            self.scores = self.scores.drop(columns=['pubication'])
            self.scores = self.scores.reindex(
                columns=self.scores.columns.tolist() + ['publication.id'
                    , 'publication.title'
                    , 'publication.doi'
                    , 'publication.PMID'
                    , 'publication.journal'
                    , 'publication.firstauthor'
                    , 'publication.date_publication'])
        if 'sample_age' in self.samples_variants.columns:
            self.samples_variants = self.samples_variants.drop(columns=['sample_age'])
        if 'sample_age.interval' in self.samples_variants.columns:
            self.samples_variants = self.samples_variants.drop(columns=['sample_age.interval'])
        if 'followup_time' in self.samples_variants.columns:
            self.samples_variants = self.samples_variants.drop(columns=['followup_time'])
        if 'followup_time.interval' in self.samples_variants.columns:
            self.samples_variants = self.samples_variants.drop(columns=['followup_time.interval'])
        if 'sample_age' in self.samples_training.columns:
            self.samples_training = self.samples_training.drop(columns=['sample_age'])
        if 'sample_age.interval' in self.samples_training.columns:
            self.samples_training = self.samples_training.drop(columns=['sample_age.interval'])
        if 'followup_time' in self.samples_training.columns:
            self.samples_training = self.samples_training.drop(columns=['followup_time'])
        if 'followup_time.interval' in self.samples_training.columns:
            self.samples_training = self.samples_training.drop(columns=['followup_time.interval'])
        for miss_column in ['sample_age.estimate_type', 'sample_age.estimate', 'sample_age.interval.type',
                            'sample_age.interval.lower', 'sample_age.interval.upper', 'sample_age.variability_type',
                            'sample_age.variability', 'sample_age.unit', 'followup_time.estimate_type',
                            'followup_time.estimate', 'followup_time.interval.type', 'followup_time.interval.lower',
                            'followup_time.interval.upper', 'followup_time.variability_type',
                            'followup_time.variability', 'followup_time.unit']:
            if miss_column not in self.samples_training.columns:
                self.samples_training[miss_column] = None
            if miss_column not in self.samples_variants.columns:
                self.samples_variants[miss_column] = None
        return

    def __str__(self):
        if self.mode == 'Fat':
            return ("Score is running in fat mode. It has 7 DataFrames with hierarchical dependencies.\n"
                    "scores:%d rows\n|\n -samples_variants: %d rows\n  |\n   -samples_variants_cohorts: %d rows\n|\n "
                    "-samples_training: %d rows"
                    "\n  |\n   -samples_training_cohorts: %d rows"
                    "\n|\n -trait_efo: %d rows"
                    "\n|\n -ancestry_distribution: %d rows" % (
                        len(self.scores), len(self.samples_variants), len(self.samples_variants_cohorts),
                        len(self.samples_training), len(self.samples_training_cohorts), len(self.trait_efo),
                        len(self.ancestry_distribution)))
        if self.mode == 'Thin':
            return ('Score is running in thin mode. It has 1 list that contains the raw data.\n raw_data: '
                    'a list of size %d.' % len(self.raw_data))

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        if isinstance(item, str) or isinstance(item, int):
            arr = [item]
        elif isinstance(item, list) or isinstance(item, tuple) or isinstance(item, range):
            arr = item
        elif isinstance(item, slice):
            start, stop, step = item.indices(len(self))
            arr = list(range(start, stop, step))
        else:
            raise TypeError('Invalid argument type：{}'.format(type(item)))
        raw_data = self.raw_data
        raw_data_dict = {}
        for j in raw_data:
            raw_data_dict[j['id']] = j
        sub_set = []
        for i in arr:
            if isinstance(i, str):
                sub_set.append(raw_data_dict[i])
            elif isinstance(i, int):
                sub_set.append(raw_data[i])
            else:
                raise TypeError('Invalid item type: {}'.format(type(i)))
        return Score(sub_set, self.mode)

    def __add__(self, other):
        if self.mode == other.mode:
            return Score(self.raw_data + other.raw_data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __sub__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set - other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Score(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __and__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set & other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return Score(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __or__(self, other):
        if self.mode == other.mode:
            and_dict = {}
            for i in self.raw_data:
                and_dict[i['id']] = i
            for j in other.raw_data:
                and_dict[j['id']] = j
            data = list(and_dict.values())
            return Score(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __xor__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            and_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                and_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
                and_dict[j['id']] = j
            sub_key = self_key_set ^ other_key_set
            data = []
            for k in sub_key:
                data.append(and_dict[k])
            return Score(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __eq__(self, other):
        if self is None or other is None:
            return self is None and other is None
        return self.raw_data == other.raw_data and self.mode == other.mode

    def __len__(self):
        return len(self.raw_data)



from pandaspgs.get_ancestry_category import get_ancestry_categories


def test_get_ancestry_categories():
    filter_by_id = get_ancestry_categories()
    assert len(filter_by_id) == 11
    assert len(filter_by_id ^ filter_by_id[0]) == 10
    assert len(filter_by_id[range(2)]) == 2
    assert len(filter_by_id[1:3]) == 2
    assert len(filter_by_id["AFR"]) == 1
    assert len(filter_by_id[0] + filter_by_id[1]) == 2
    assert len(filter_by_id - filter_by_id[1]) == 10
    assert len(filter_by_id[0] & filter_by_id) == 1
    assert len(filter_by_id | filter_by_id[0]) == 11
    assert len(filter_by_id[0:4] | filter_by_id[4]) == 5


from pandas import DataFrame, json_normalize, set_option, Series

set_option('display.max_columns', None)
set_option('display.width', 1000)
set_option('display.colheader_justify', 'center')
set_option('display.precision', 3)
import numpy


class PerformanceMetric:
    """
    An object that stores data of type PerformanceMetric. To understand the significance of each column of the DataFrame. Please visit "PerformanceMetric" in [PGS Catalog Documentation](https://www.pgscatalog.org/rest/) for details.

    Attributes:
         raw_data: list. Convert from obtained JSON data
         performance_metrics: DataFrame. It only exists if the parameter mode of constructor is Fat.
         samples: DataFrame. It only exists if the parameter mode of constructor is Fat.
         cohorts: DataFrame. It only exists if the parameter mode of constructor is Fat.
         effect_sizes: DataFrame. It only exists if the parameter mode of constructor is Fat.
         class_acc: DataFrame. It only exists if the parameter mode of constructor is Fat.
         othermetrics: DataFrame. It only exists if the parameter mode of constructor is Fat.
         mode: Fat or Thin. Specifies the mode of the returned object.

    ```python
    from pandaspgs.get_performance import get_performances

    ch = get_performances(ppm_id='PPM000001')
    ch
    ch.raw_data
    ch.mode
    ch.performance_metrics
    ch.samples
    ch.cohorts
    ch.effect_sizes
    ch.class_acc
    ch.othermetrics
    ```
    Subset object s by either identifier or position
    ```python
    all_df = get_performance()
    all_df[0].performance_metrics
    all_df[0:3].performance_metrics
    all_df['PPM000001'].performance_metrics
    all_df[('PPM000001','PPM000002','PPM000003')].performance_metrics
    ```
    Objects can be manipulated like sets in the mathematical sense.
    ```python
    one = get_cohorts(ppm_id='PPM000001')
    two = get_cohorts(ppm_id='PPM000002')
    three = get_cohorts(ppm_id='PPM000003')
    one_and_two = one+two
    two_and_three = two+three
    only_one = one_and_two - two_and_three
    only_two = one_and_two & two_and_three
    one_and_two_and_three = one_and_two | two_and_three
    one_and_three = one_and_two ^ two_and_three
    ```
    """
    def __init__(self, data: list = [], mode: str = "Fat"):
        """
        Args:
            data: Raw JSON data.
            mode: Fat or Thin. Specifies the mode of the object.
        """
        if data is None:
            data = []
        if mode not in ['Thin', "Fat"]:
            raise Exception("Mode must be Fat or Thin")
        self.raw_data = data
        self.mode = mode
        if mode == "Thin":
            return
        if data is None or len(data) == 0:
            self.performance_metrics = DataFrame(
                columns=['id', 'associated_pgs_id', 'phenotyping_reported', 'publication.id', 'publication.title',
                         'publication.doi', 'publication.PMID', 'publication.journal', 'publication.firstauthor',
                         'publication.date_publication', 'sampleset.id', 'covariates',
                         'performance_comments'])
            self.samples = DataFrame(
                columns=['id', 'performance_metric_id', 'sample_number', 'sample_cases', 'sample_controls',
                         'sample_percent_male', 'sample_age.estimate_type', 'sample_age.estimate',
                         'sample_age.interval.type', 'sample_age.interval.lower', 'sample_age.interval.upper',
                         'sample_age.variability_type', 'sample_age.variability', 'sample_age.unit',
                         'phenotyping_free', 'followup_time.estimate_type', 'followup_time.estimate',
                         'followup_time.interval.type', 'followup_time.interval.lower', 'followup_time.interval.upper',
                         'followup_time.variability_type', 'followup_time.variability', 'followup_time.unit',
                         'ancestry_broad', 'ancestry_free', 'ancestry_country', 'ancestry_additional',
                         'source_GWAS_catalog', 'source_PMID', 'source_DOI', 'cohorts_additional'])
            self.cohorts = DataFrame(
                columns=['performance_metric_id', 'sample_id', 'name_short', 'name_full', 'name_others'])
            self.effect_sizes = DataFrame(
                columns=['performance_metric_id', 'name_long', 'name_short', 'estimate', 'ci_lower', 'ci_upper', 'se',
                         'unit'])
            self.class_acc = DataFrame(
                columns=['performance_metric_id', 'name_long', 'name_short', 'estimate', 'ci_lower', 'ci_upper', 'se',
                         'unit'])
            self.othermetrics = DataFrame(
                columns=['performance_metric_id', 'name_long', 'name_short', 'estimate', 'ci_lower', 'ci_upper', 'se',
                         'unit'])
            return
        datas = json_normalize(data=data, max_level=1)
        datas['sampleset.samples'] = datas['sampleset.samples'].map(lambda x: x == [])
        datas['performance_metrics.effect_sizes'] = datas['performance_metrics.effect_sizes'].map(lambda x: x == [])
        datas['performance_metrics.class_acc'] = datas['performance_metrics.class_acc'].map(lambda x: x == [])
        datas['performance_metrics.othermetrics'] = datas['performance_metrics.othermetrics'].map(lambda x: x == [])
        self.performance_metrics = json_normalize(data=data, max_level=1).drop(
            columns=['sampleset.samples', 'performance_metrics.effect_sizes',
                     'performance_metrics.class_acc',
                     'performance_metrics.othermetrics'])
        if not datas['sampleset.samples'].all():
            self.samples = json_normalize(data=data, record_path=['sampleset', 'samples'], meta=['id'])
            self.samples['performance_metric_id'] = self.samples['id']
            self.samples['id'] = Series(data=range(0, len(self.samples)))
            cohort = self.samples[['id', 'performance_metric_id', 'cohorts']].copy()
            self.samples = self.samples.drop(columns=['cohorts'])
            cohort['sample_id'] = cohort['id']
            cohort['cohorts'] = cohort['cohorts'].apply(lambda x: x if len(x) > 0 else numpy.nan)
            cohort = cohort.dropna()
            cohort = cohort.explode('cohorts')
            if len(cohort) == 0:
                self.cohorts = DataFrame(
                    columns=['performance_metric_id', 'sample_id', 'name_short', 'name_full', 'name_others'])
            cohort[['name_short', 'name_full', 'name_others']] = cohort['cohorts'].apply(
                lambda x: Series(data=[x['name_short'], x['name_full'], x['name_others']]))
            cohort = cohort.drop(columns=['id', 'cohorts'])
            self.cohorts = cohort

        else:
            self.samples = DataFrame(
                columns=['id', 'performance_metric_id', 'sample_number', 'sample_cases', 'sample_controls',
                         'sample_percent_male', 'sample_age.estimate_type', 'sample_age.estimate',
                         'sample_age.interval.type', 'sample_age.interval.lower', 'sample_age.interval.upper',
                         'sample_age.variability_type', 'sample_age.variability', 'sample_age.unit',
                         'phenotyping_free', 'followup_time.estimate_type', 'followup_time.estimate',
                         'followup_time.interval.type', 'followup_time.interval.lower', 'followup_time.interval.upper',
                         'followup_time.variability_type', 'followup_time.variability', 'followup_time.unit',
                         'ancestry_broad', 'ancestry_free', 'ancestry_country', 'ancestry_additional',
                         'source_GWAS_catalog', 'source_PMID', 'source_DOI', 'cohorts_additional'])
            self.cohorts = DataFrame(
                columns=['performance_metric_id', 'sample_id', 'name_short', 'name_full', 'name_others'])
        if not datas['performance_metrics.effect_sizes'].all():
            self.effect_sizes = json_normalize(data=data, record_path=['performance_metrics', 'effect_sizes'],
                                               meta=['id'])
            self.effect_sizes['performance_metric_id'] = self.effect_sizes['id']
            self.effect_sizes = self.effect_sizes.drop(columns=['id'])
        else:
            self.effect_sizes = DataFrame(
                columns=['performance_metric_id', 'name_long', 'name_short', 'estimate', 'ci_lower', 'ci_upper', 'se'])
        if not datas['performance_metrics.class_acc'].all():
            self.class_acc = json_normalize(data=data, record_path=['performance_metrics', 'class_acc'], meta=['id'])
            self.class_acc['performance_metric_id'] = self.class_acc['id']
            self.class_acc = self.class_acc.drop(columns=['id'])
        else:
            self.class_acc = DataFrame(
                columns=['performance_metric_id', 'name_long', 'name_short', 'estimate', 'ci_lower', 'ci_upper', 'se'])
        if not datas['performance_metrics.othermetrics'].all():
            self.othermetrics = json_normalize(data=data, record_path=['performance_metrics', 'othermetrics'],
                                               meta=['id'])
            self.othermetrics['performance_metric_id'] = self.othermetrics['id']
            self.othermetrics = self.othermetrics.drop(columns=['id'])
        else:
            self.othermetrics = DataFrame(
                columns=['performance_metric_id', 'name_long', 'name_short', 'estimate', 'ci_lower', 'ci_upper', 'se'])
        if 'publication' in self.performance_metrics.columns:
            self.performance_metrics = self.performance_metrics.drop(columns=['pubication'])
            self.performance_metrics = self.performance_metrics.reindex(
                columns=self.performance_metrics.columns.tolist() + ['publication.title',
                                                                     'publication.doi', 'publication.PMID',
                                                                     'publication.journal',
                                                                     'publication.firstauthor',
                                                                     'publication.date_publication'])
        if 'sample_age' in self.samples.columns:
            self.samples = self.samples.drop(columns=['sample_age'])
            if 'sample_age.estimate_type' not in self.samples.columns:
                self.samples = self.samples.reindex(
                    columns=self.samples.columns.tolist() + ['sample_age.estimate_type', 'sample_age.estimate',
                                                             'sample_age.interval.type', 'sample_age.interval.lower',
                                                             'sample_age.interval.upper',
                                                             'sample_age.variability_type', 'sample_age.variability',
                                                             'sample_age.unit']
            )
        if 'followup_time' in self.samples.columns:
            self.samples = self.samples.drop(columns=['followup_time'])
            if 'followup_time.estimate_type' not in self.samples.columns:
                self.samples = self.samples.reindex(
                    columns=self.samples.columns.tolist() + ['followup_time.estimate_type', 'followup_time.estimate',
                                                             'followup_time.interval.type',
                                                             'followup_time.interval.lower',
                                                             'followup_time.interval.upper',
                                                             'followup_time.variability_type',
                                                             'followup_time.variability',
                                                             'followup_time.unit'])
        if 'unit' not in self.effect_sizes.columns:
            self.effect_sizes['unit'] = None
        if 'unit' not in self.class_acc.columns:
            self.class_acc['unit'] = None
        if 'unit' not in self.othermetrics.columns:
            self.othermetrics['unit'] = None
        return

    def __str__(self):
        if self.mode == 'Fat':
            return ("PerformanceMetric is running in fat mode. It has 6 DataFrames with hierarchical dependencies.\n"
                    "performance_metrics:"
                    "%d rows\n|\n -samples: %d rows\n  |\n   -cohorts: %d rows\n|\n -effect_sizes: %d rows"
                    "\n|\n -class_acc: %d rows"
                    "\n|\n -othermetrics: %d rows" % (
                        len(self.performance_metrics), len(self.samples), len(self.cohorts),
                        len(self.effect_sizes), len(self.class_acc), len(self.othermetrics)))
        if self.mode == 'Thin':
            return ('PerformanceMetrics is running in thin mode. It has 1 list that contains the raw data.\nraw_data: '
                    'a list of size %d.' % len(self.raw_data))

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        if isinstance(item, str) or isinstance(item, int):
            arr = [item]
        elif isinstance(item, list) or isinstance(item, tuple) or isinstance(item, range):
            arr = item
        elif isinstance(item, slice):
            start, stop, step = item.indices(len(self))
            arr = list(range(start, stop, step))
        else:
            raise TypeError('Invalid argument type：{}'.format(type(item)))
        raw_data = self.raw_data
        raw_data_dict = {}
        for j in raw_data:
            raw_data_dict[j['id']] = j
        sub_set = []
        for i in arr:
            if isinstance(i, str):
                sub_set.append(raw_data_dict[i])
            elif isinstance(i, int):
                sub_set.append(raw_data[i])
            else:
                raise TypeError('Invalid item type: {}'.format(type(i)))
        return PerformanceMetric(sub_set, self.mode)

    def __add__(self, other):
        if self.mode == other.mode:
            return PerformanceMetric(self.raw_data + other.raw_data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __sub__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set - other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return PerformanceMetric(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __and__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            self_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                self_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
            sub_key = self_key_set & other_key_set
            data = []
            for k in sub_key:
                data.append(self_dict[k])
            return PerformanceMetric(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __or__(self, other):
        if self.mode == other.mode:
            and_dict = {}
            for i in self.raw_data:
                and_dict[i['id']] = i
            for j in other.raw_data:
                and_dict[j['id']] = j
            data = list(and_dict.values())
            return PerformanceMetric(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __xor__(self, other):
        if self.mode == other.mode:
            self_key_set = set()
            and_dict = {}
            other_key_set = set()
            for i in self.raw_data:
                self_key_set.add(i['id'])
                and_dict[i['id']] = i
            for j in other.raw_data:
                other_key_set.add(j['id'])
                and_dict[j['id']] = j
            sub_key = self_key_set ^ other_key_set
            data = []
            for k in sub_key:
                data.append(and_dict[k])
            return PerformanceMetric(data, self.mode)
        else:
            raise Exception("Please input the same mode")

    def __eq__(self, other):
        if self is None or other is None:
            return self is None and other is None
        return self.raw_data == other.raw_data and self.mode == other.mode

    def __len__(self):
        return len(self.raw_data)



from typing import List, Dict
from pandaspgs.client import get_release
from pandaspgs.release import Release
import re


def get_releases(date: str = 'all', cached: bool = True, mode: str = 'Fat') -> Release:
    """
    Get Release data from the server.

    Args:
        date: PGS Catalog release date (format YYYY-MM-DD) or 'latest' or 'all'
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A Release object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_release import get_releases

    ch = get_releases(date="2024-01-26")
    ```
    """
    if date == 'all':
        return Release(get_release('https://www.pgscatalog.org/rest/release/all', cached=cached), mode)
    if date == 'latest':
        return Release(get_release('https://www.pgscatalog.org/rest/release/current', cached=cached), mode)
    if date is None:
        raise Exception("Date can't be None.")
    if re.match('\\d{4}-\\d{2}-\\d{2}$', date) is None:
        raise Exception('The format of the date must be YYYY-MM-DD.')
    return Release(get_release('https://www.pgscatalog.org/rest/release/%s' % date, cached=cached), mode)






import requests
import json
import progressbar
from typing import List, Dict
from requests.adapters import HTTPAdapter
from cachetools import TTLCache

fields = ['Score', 'Publication', 'Trait', 'Trait_category', 'Performance', 'Cohort', 'Sample_set', 'Release',
          'Ancestry_category']
fields_and_all = ['All', 'Score', 'Publication', 'Trait', 'Trait_category', 'Performance', 'Cohort', 'Sample_set',
                  'Release', 'Ancestry_category']
publication_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
score_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
trait_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
trait_category_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
performance_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
cohort_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
sample_set_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
release_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)
ancestry_category_cache = TTLCache(maxsize=1024*1024, ttl=60 * 60 * 24)


def get_ancestry_category(url: str, cached=True) -> List[Dict]:
    raw_dict = get_data(url, cache_impl=ancestry_category_cache, cached=cached)[0]
    dict_list = []
    for key in raw_dict:
        raw_dict[key]['symbols'] = key
        dict_list.append(raw_dict[key])
    return dict_list


def get_publication(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=publication_cache, cached=cached)


def get_score(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=score_cache, cached=cached)


def get_trait(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=trait_cache, cached=cached)


def get_trait_category(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=trait_category_cache, cached=cached)


def get_performance(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=performance_cache, cached=cached)


def get_cohort(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=cohort_cache, cached=cached)


def get_sample_set(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=sample_set_cache, cached=cached)


def get_release(url: str, cached=True) -> List[Dict]:
    return get_data(url, cache_impl=release_cache, cached=cached)


def clear_cache(field: str = 'All') -> None:
    """
    Clear some or all of the cache.

    Args:
        field: It can be one of the following: 'All', 'Score', 'Publication', 'Trait', 'Trait_category', 'Performance', 'Cohort', 'Sample_set', 'Release', 'Ancestry_category'

    Returns:
        None

    ```Python
    from pandaspgs.get_publication import get_publications
    from pandaspgs.client import clear_cache

    # Clear all caches.
    clear_cache('All')
    pub = get_publications()
    # Clear the cache used by get_publications()
    clear_cache('Publication')
    pub = get_publications()
    ```

    """
    if field not in fields_and_all:
        raise Exception('The field must one of %s' % str(fields_and_all))
    if field == 'All':
        for s in fields:
            eval(s.lower() + '_cache.clear()')
    else:
        eval(field.lower() + '_cache.clear()')


def get_data(url: str, cache_impl=None, cached=True) -> List[Dict]:
    with requests.Session() as s:
        s.mount('https://', HTTPAdapter(max_retries=5))
        if url in cache_impl and cached:
            r = cache_impl[url]
        else:
            r = s.get(url)
        if r.status_code == 200:
            if len(json.loads(r.text)) == 0:
                return []
            cache_impl[url] = r
            parsed_data = json.loads(r.text)
            if parsed_data.get('results') is not None:
                results_list = parsed_data.get('results')
                if parsed_data.get('next') is not None:
                    bar = progressbar.ProgressBar(max_value=parsed_data.get('count')).start()
                    progress = 50
                    bar.update(progress)
                    next_url = parsed_data.get('next')
                    while next_url is not None:
                        if next_url in cache_impl and cached:
                            r = cache_impl[next_url]
                        else:
                            r = s.get(next_url)
                            if r.status_code == 200:
                                cache_impl[next_url] = r
                            elif r.status_code == 404:
                                return []
                            else:
                                raise Exception('The request for %s failed: response code was %d, content was \n%s' % (next_url, r.status_code, r.text))
                        parsed_data = json.loads(r.text)
                        results_list.extend(parsed_data.get('results'))
                        progress = progress + parsed_data.get('size')
                        bar.update(progress)
                        next_url = parsed_data.get('next')
                    bar.finish()
                return results_list
            else:
                return [parsed_data]
        elif r.status_code == 404:
            return []
        else:
            raise Exception('The request for %s failed: response code was %d, content was \n%s' % (url, r.status_code, r.text))


def ask_yes_no_question(question: str) -> str:
    yes_no_answer = ""
    while yes_no_answer != "YES" and yes_no_answer != "NO":
        yes_no_answer = input(question).upper()
    return yes_no_answer



from pandaspgs import Cohort
from pandaspgs.ancestry_category import AncestryCategory
from pandaspgs.performance import PerformanceMetric
from pandaspgs.publication import Publication
from pandaspgs.release import Release
from pandaspgs.sample_set import SampleSet
from pandaspgs.score import Score
from pandaspgs.trait import Trait
from pandaspgs.trait_category import TraitCategory


def bind(
        a: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory,
        b: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory:
    """
    Binds together PGS objects of the same object. Note that bind() preserves duplicates whereas union() does not.

    Args:
        a: An object of the pandasPGS custom class.
        b: An object of the same type as a.

    Returns:
        An object of the same type as a.
    ```python
    from pandaspgs.get_cohort import get_cohorts
    from pandaspgs.set_operation import bind


    a = get_cohorts(cohort_symbol='100-plus')
    b = get_cohorts(cohort_symbol='23andMe')
    c = bind(a,b)
    ```

    """
    return a + b


def intersect(
        a: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory,
        b: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory:
    """
    Returns the data common to both A and B, with no repetitions

    Args:
        a: An object of the pandasPGS custom class.
        b:An object of the same type as a.

    Returns:
        An object of the same type as a.
    ```python
    from pandaspgs.get_cohort import get_cohorts
    from pandaspgs.set_operation import intersect


    a = get_cohorts(cohort_symbol='100-plus')
    b = get_cohorts(cohort_symbol='23andMe')
    c = intersect(a,b)
    ```

    """
    return a & b


def set_diff(
        a: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory,
        b: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory:
    """
    returns the data in A that is not in B, with no repetitions

    Args:
        a: An object of the pandasPGS custom class.
        b:An object of the same type as a.

    Returns:
        An object of the same type as a.
    ```python
    from pandaspgs.get_cohort import get_cohorts
    from pandaspgs.set_operation import set_diff


    a = get_cohorts(cohort_symbol='23andMe')
    b = get_cohorts(cohort_symbol='23andMe')
    c = set_diff(a,b)
    ```

    """
    return a - b


def set_xor(
        a: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory,
        b: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory:
    """
    returns the data of A and B that are not in their intersection (the symmetric difference), with no repetitions

    Args:
        a: An object of the pandasPGS custom class.
        b:An object of the same type as a.

    Returns:
        An object of the same type as a.
    ```python
    from pandaspgs.get_cohort import get_cohorts
    from pandaspgs.set_operation import set_xor


    a = get_cohorts(cohort_symbol='100-plus')
    b = get_cohorts(cohort_symbol='23andMe')
    c = set_xor(a,b)
    ```

    """
    return a ^ b


def union(
        a: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory,
        b: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory:
    """
    returns the combined data from A and B with no repetitions

    Args:
        a: An object of the pandasPGS custom class.
        b:An object of the same type as a.

    Returns:
        An object of the same type as a.
    ```python
    from pandaspgs.get_cohort import get_cohorts
    from pandaspgs.set_operation import union


    a = get_cohorts(cohort_symbol='100-plus')
    b = get_cohorts(cohort_symbol='23andMe')
    c = union(a,b)
    ```

    """
    return a | b


def set_equal(
        a: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory,
        b: AncestryCategory | Cohort | PerformanceMetric | Publication | Release | SampleSet | Score | Trait | TraitCategory) -> bool:
    """
    Check if the raw data of a and b are equal

    Args:
        a: An object of the pandasPGS custom class.
        b:An object of the same type as a.

    Returns:
        An object of the same type as a.
    ```python
    from pandaspgs.get_cohort import get_cohorts
    from pandaspgs.set_operation import set_equal


    a = get_cohorts(cohort_symbol='100-plus')
    b = get_cohorts(cohort_symbol='23andMe')
    c = set_equal(a,b)
    ```

    """
    return a == b



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pandaspgs",
    version="0.1.0",
    author="Cao Tianze",
    author_email="hnrcao@qq.com",
    description="A Python package for easy retrieval of PGS Catalog data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tianzelab/pandaspgs",
    project_urls={
        "Bug Tracker": "https://github.com/tianzelab/pandaspgs/issues",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=['pandaspgs'],
    python_requires=">=3.11",
    install_requires=['pandas>=2.1.4', 'requests>=2.31.0', 'progressbar2>=4.4.2', 'cachetools>=5.3.3'],
    license="MIT",
    keywords=['pgs', 'genomics', 'snp', 'bioinformatics','pandas'],
    package_data={
        "": ["*.csv","*.txt"]
    }
)



from typing import List, Dict
from pandaspgs.client import get_score
from pandas import DataFrame, json_normalize, set_option, Series, concat
from pandaspgs.score import Score


def get_scores(pgs_id: str = None, pgp_id: str = None, pmid: int = None, trait_id: str = None, cached: bool = True,
               mode: str = 'Fat') -> Score:
    """
    Get Score data from the server.

    Args:
        pgs_id: Polygenic Score ID
        pgp_id: PGS Catalog Publication ID (PGP)
        pmid: PubMed ID (without the prefix "PMID:")
        trait_id: Ontology ID (e.g. from EFO, HP or MONDO) with the format "EFO_XXXX"
        cached: Whether or not to try to get data from the cache.
        mode: Fat or Thin. Specifies the mode of the returned object.


    Returns:
        A Score object. Attributes of type DataFrame have hierarchical dependencies.

    ```Python
    from pandaspgs.get_score import get_scores

    ch = get_scores(pgs_id='PGS000001')
    ```
    """
    if pgs_id is None and pgp_id is None and pmid is None and trait_id is None:
        return Score(get_score('https://www.pgscatalog.org/rest/score/all', cached=cached), mode)
    by_pgs_id = None
    by_other = None
    if pgs_id is not None:
        by_pgs_id = get_score('https://www.pgscatalog.org/rest/score/%s' % pgs_id, cached=cached)
    if pgp_id is not None or pmid is not None or trait_id is not None:
        query_str = []
        if pgp_id is not None:
            query_str.append('pgp_id=%s' % pgp_id)
        if pmid is not None:
            query_str.append('pmid=%d' % pmid)
        if trait_id is not None:
            query_str.append('trait_id=%s' % trait_id)
        by_other = get_score('https://www.pgscatalog.org/rest/score/search?%s' % '&'.join(query_str))
    if pgs_id is None:
        return Score(by_other, mode)
    if pgp_id is None and pmid is None and trait_id is None:
        return Score(by_pgs_id, mode)
    other_set = set()
    pgs_id_dict = {}
    for single in by_pgs_id:
        pgs_id_dict[single['id']] = single
    pgs_id_set = pgs_id_dict.keys()
    for single in by_other:
        other_set.add(single['id'])
    intersection = pgs_id_set & other_set
    result = []
    for id in intersection:
        result.append(pgs_id_dict[id])
    return Score(result, mode)





from pandaspgs.get_trait import get_trait_categories
from pandaspgs.get_trait import get_traits
import pytest


def test_get_trait_categories():
    categories = get_trait_categories(mode='Thin')
    categories
    categories = get_trait_categories()
    categories
    assert len(categories) == 17
    assert len(categories[1:3]) == 2
    assert len(categories[range(2)]) == 2
    assert len(categories ^ categories[0]) == 16
    assert len(categories[0] + categories[1]) == 2
    assert len(categories - categories[1]) == 16
    assert len(categories & categories[1]) == 1
    assert len(categories[0] | categories[1]) == 2


def test_get_traits():
    filter_get_trait = get_traits(trait_id='EFO_0000305', mode='Thin')
    filter_get_trait
    filter_get_trait = get_traits(trait_id='EFO_0000305')
    filter_get_trait
    assert len(filter_get_trait) == 1
    assert filter_get_trait.traits.size == 4 * 1
    assert filter_get_trait.trait_categories.size == 2 * 1
    assert filter_get_trait.trait_synonyms.size == 2 * 9
    assert filter_get_trait.associated_pgs_ids.size == 2 * 103
    assert filter_get_trait.child_associated_pgs_ids.size == 2 * 17
    assert filter_get_trait.trait_mapped_terms.size == 2 * 8
    filter_get_trait_a = get_traits(trait_id='EFO_0001645')
    assert len(filter_get_trait_a) == 1
    filter_get_trait_b = get_traits(term='Alzheimer')
    assert len(filter_get_trait_b) == 4
    filter_get_trait_c = get_traits(term='Neurological disorder')
    assert len(filter_get_trait_c) == 102
    filter_get_trait_d = get_traits(trait_id="EFO_0005782", term='Neurological disorder')
    assert len(filter_get_trait_d) == 1
    filter_get_trait_e = get_traits()
    assert len(filter_get_trait_e) == 1180
    filter_get_trait_f = get_traits(term='Alzheimer', exact=False)
    assert len(filter_get_trait_f) == 4
    filter_get_trait_g = get_traits(trait_id="EFO_0005782", term='Neurological disorder', exact=False)
    assert len(filter_get_trait_g) == 1
    assert filter_get_trait[0] == filter_get_trait
    assert filter_get_trait['EFO_0000305'] == filter_get_trait
    assert len(filter_get_trait_c[1:3]) == 2
    assert len(filter_get_trait_c[range(2)]) == 2
    assert filter_get_trait[0] == filter_get_trait['EFO_0000305']
    assert len(filter_get_trait_c ^ filter_get_trait_c[0]) == len(filter_get_trait_c[1:102])
    assert len(filter_get_trait_a + filter_get_trait_c) == 103
    assert len(filter_get_trait_c - filter_get_trait_c[0]) == 101
    assert len(filter_get_trait_c & filter_get_trait_c[0]) == 1
    assert len(filter_get_trait_c | filter_get_trait_a) == 103
    with pytest.raises(Exception):
        get_traits(trait_id='EFO_0000305', exact=False)



from pandaspgs import get_releases


def test_get_releases():
    current = get_releases()
    assert len(current) == 80
    all_release = get_releases(date='all')
    assert len(all_release) == 80
    some_release = get_releases(date='2023-10-17')
    assert len(some_release) == 1
    assert len(all_release[1:3]) == 2
    assert len(all_release[2]) == 1
    assert len(all_release['2023-10-17']) == 1
    assert len(all_release[range(2)]) == 2
    assert len(all_release ^ all_release[0]) == 79
    assert len(all_release[0] + all_release[1]) == 2
    assert len(all_release - all_release[0]) == 79
    assert len(all_release & all_release[1]) == 1
    assert len(all_release[0] | all_release[1]) == 2
